"""
Code for commonalities between "X for" objects that compute analytic properties
of networks like OutputShapeFor and ReceptiveFieldFor
"""
import ubelt as ub
from collections import OrderedDict


class Hidden(OrderedDict, ub.NiceRepr):
    """ Object for storing hidden states of analystic computation """

    def __nice__(self):
        return ub.repr2(self, nl=0)

    def __str__(self):
        return ub.NiceRepr.__str__(self)

    def __repr__(self):
        return ub.NiceRepr.__repr__(self)

    def __setitem__(self, key, value):
        if getattr(value, 'hidden', None) is not None:
            # When setting a value to an OutputShape object, if that object has
            # a hidden shape, then use that instead.
            value = value.hidden
        return OrderedDict.__setitem__(self, key, value)

    def shallow(self, n=1):
        """
        Grabs only the shallowest n layers of hidden shapes
        """
        if n == 0:
            last = self
            while hasattr(last, 'shallow'):
                values = list(last.values())
                if len(values):
                    last = values[-1]
                else:
                    break
            return last
        else:
            output = OrderedDict()
            for key, value in self.items():
                # if isinstance(value, HiddenShapes):
                if hasattr(value, 'shallow'):
                    value = value.shallow(n - 1)
                output[key] = value
            return output


class OutputFor(object):
    """
    Analytic base / identity class
    """
    def __init__(self, func):
        self.func = func

    def __call__(self, *args, **kw):
        return self.func(*args, **kw)


class Output(object):
    """
    Analytic base / identity class
    """
    @classmethod
    def coerce(cls, data=None, hidden=None):
        return data


class ForwardFor(OutputFor):
    """
    Analytic version of forward functions
    """
    def __init__(self, func):
        self.func = func

    def __call__(self, *args, **kw):
        return self.func(*args, **kw)

    @staticmethod
    def getitem(arr):
        """
        Wraps getitem calls

        Example:
            >>> import torch
            >>> arr = torch.rand(2, 16, 2, 2)
            >>> result = ForwardFor.getitem(arr)[:, 0:4]
            >>> assert result.shape == (2, 4, 2, 2)
        """
        return _ForwardGetItem(arr)

    @staticmethod
    def view(arr, *args):
        """
        Wraps view calls

        Example:
            >>> import torch
            >>> arr = torch.rand(2, 16, 2, 2)
            >>> result = ForwardFor.view(arr, -1)
        """
        return arr.view(*args)

    @staticmethod
    def shape(arr):
        """
        Wraps shape calls

        Example:
            >>> import torch
            >>> arr = torch.rand(2, 16, 2, 2)
            >>> result = ForwardFor.shape(arr)
        """
        return arr.shape

    @staticmethod
    def add(arr1, arr2):
        return arr1 + arr2

    @staticmethod
    def mul(arr1, arr2):
        return arr1 * arr2

    @staticmethod
    def sub(arr1, arr2):
        return arr1 - arr2

    @staticmethod
    def div(arr1, arr2):
        return arr1 - arr2


class _ForwardGetItem(object):
    def __init__(self, inp):
        self.inp = inp

    def __getitem__(self, slices):
        return self.inp.__getitem__(slices)



"""
mkinit netharn.models
"""
# flake8: noqa

__DYNAMIC__ = False
if __DYNAMIC__:
    from mkinit import dynamic_init
    exec(dynamic_init(__name__))
else:
    # <AUTOGEN_INIT>
    from netharn.models import densenet
    from netharn.models import descriptor_network
    from netharn.models import dual_path_net
    from netharn.models import resnet
    from netharn.models import toynet
    from netharn.models import yolo2

    from netharn.models.densenet import (DenseNet,)
    from netharn.models.descriptor_network import (DescriptorNetwork,)
    from netharn.models.dual_path_net import (Bottleneck, DPN, DPN26, DPN92,)
    from netharn.models.resnet import (ResNet,)
    from netharn.models.toynet import (ToyNet1d, ToyNet2d,)

    __all__ = ['Bottleneck', 'DPN', 'DPN26', 'DPN92', 'DenseNet',
               'DescriptorNetwork', 'ResNet', 'ToyNet1d', 'ToyNet2d', 'densenet',
               'descriptor_network', 'dual_path_net', 'resnet', 'toynet', 'yolo2']



"""
Simple dataset for loading the VOC 2007 object detection dataset without extra
bells and whistles. Simply loads the images, boxes, and class labels and
resizes images to a standard size.
"""
from os.path import exists
from os.path import join
import re
import torch
import glob
import ubelt as ub
import numpy as np
from netharn.data import collate
import torch.utils.data as torch_data


class VOCDataset(torch_data.Dataset, ub.NiceRepr):
    """
    Example:
        >>> # xdoc: +REQUIRES(--voc)
        >>> assert len(VOCDataset(split='train', years=[2007])) == 2501
        >>> assert len(VOCDataset(split='test', years=[2007])) == 4952
        >>> assert len(VOCDataset(split='val', years=[2007])) == 2510
        >>> assert len(VOCDataset(split='trainval', years=[2007])) == 5011

        >>> assert len(VOCDataset(split='train', years=[2007, 2012])) == 8218
        >>> assert len(VOCDataset(split='test', years=[2007, 2012])) == 4952
        >>> assert len(VOCDataset(split='val', years=[2007, 2012])) == 8333

    Example:
        >>> # xdoc: +REQUIRES(--voc)
        >>> years = [2007, 2012]
        >>> self = VOCDataset()
        >>> for i in range(10):
        ...     a, bc = self[i]
        ...     #print(bc[0].shape)
        ...     print(bc[1].shape)
        ...     print(a.shape)

    Example:
        >>> # xdoc: +REQUIRES(--voc)
        >>> self = VOCDataset(split='trainval', years=[2007, 2012])
        >>> test = VOCDataset(split='test', years=[2007])

        >>> self1 = VOCDataset(split='trainval', years=[2012])
        >>> self2 = VOCDataset(split='trainval', years=[2007])


        from netharn import util
        util.qtensure()
        self1.show_image(198)
        self2.show_image(300)
    """
    def __init__(self, devkit_dpath=None, split='train', years=[2007, 2012]):
        if devkit_dpath is None:
            # ub.expandpath('~/data/VOC/VOCdevkit')
            devkit_dpath = self.ensure_voc_data(years=years)

        self.devkit_dpath = devkit_dpath
        self.years = years

        # determine train / test splits
        self.gpaths = []
        self.apaths = []
        if split == 'test':
            assert 2007 in years, 'test set is hacked to be only 2007'
            gps, aps = self._read_split_paths('test', 2007)
            self.gpaths += gps
            self.apaths += aps
        else:
            for year in sorted(years):
                gps, aps = self._read_split_paths(split, year)
                self.gpaths += gps
                self.apaths += aps

        self.label_names = ('aeroplane', 'bicycle', 'bird', 'boat', 'bottle',
                            'bus', 'car', 'cat', 'chair', 'cow', 'diningtable',
                            'dog', 'horse', 'motorbike', 'person',
                            'pottedplant', 'sheep', 'sofa', 'train',
                            'tvmonitor')
        self._class_to_ind = ub.invert_dict(dict(enumerate(self.label_names)))
        self.base_wh = [416, 416]

        self.num_classes = len(self.label_names)

        import os
        hashid = ub.hash_data(list(map(os.path.basename, self.gpaths)))
        yearid = '_'.join(map(str, years))
        self.input_id = 'voc_{}_{}_{}'.format(yearid, split, hashid)

    def _read_split_paths(self, split, year):
        """
        split = 'train'
        self = VOCDataset('test')
        year = 2007
        year = 2012
        """
        split_idstrs = []
        data_dpath = join(self.devkit_dpath, 'VOC{}'.format(year))
        split_dpath = join(data_dpath, 'ImageSets', 'Main')
        pattern = join(split_dpath, '*_' + split + '.txt')
        for p in sorted(glob.glob(pattern)):
            rows = [list(re.split(' +', t)) for t in ub.readfrom(p).split('\n') if t]
            # code = -1 if the image does not contain the object
            # code = 1 if the image contains at least one instance
            # code = 0 if the image contains only hard instances of the object
            idstrs = [idstr for idstr, code in rows if int(code) == 1]
            split_idstrs.extend(idstrs)
        split_idstrs = sorted(set(split_idstrs))

        image_dpath = join(data_dpath, 'JPEGImages')
        annot_dpath = join(data_dpath, 'Annotations')
        gpaths = [join(image_dpath, '{}.jpg'.format(idstr))
                  for idstr in split_idstrs]
        apaths = [join(annot_dpath, '{}.xml'.format(idstr))
                  for idstr in split_idstrs]
        return gpaths, apaths
        # for p in gpaths:
        #     assert exists(p)
        # for p in apaths:
        #     assert exists(p)
        # return split_idstrs

    @classmethod
    def ensure_voc_data(VOCDataset, dpath=None, force=False, years=[2007, 2012]):
        """
        Download the Pascal VOC 2007 data if it does not already exist.

        CommandLine:
            python -m netharn.data.voc VOCDataset.ensure_voc_data

        Example:
            >>> # SCRIPT
            >>> # xdoc: +REQUIRES(--voc)
            >>> from netharn.data.voc import *  # NOQA
            >>> VOCDataset.ensure_voc_data()
        """
        if dpath is None:
            dpath = ub.expandpath('~/data/VOC')
        devkit_dpath = join(dpath, 'VOCdevkit')
        # if force or not exists(devkit_dpath):
        ub.ensuredir(dpath)

        fpath1 = ub.grabdata('http://host.robots.ox.ac.uk/pascal/VOC/voc2007/VOCdevkit_08-Jun-2007.tar', dpath=dpath)
        if force or not exists(join(dpath, 'VOCdevkit', 'VOCcode')):
            ub.cmd('tar xvf "{}" -C "{}"'.format(fpath1, dpath), verbout=1)

        if 2007 in years:
            # VOC 2007 train+validation data
            fpath2 = ub.grabdata('http://host.robots.ox.ac.uk/pascal/VOC/voc2007/VOCtrainval_06-Nov-2007.tar', dpath=dpath)
            if force or not exists(join(dpath, 'VOCdevkit', 'VOC2007', 'ImageSets', 'Main', 'bird_trainval.txt')):
                ub.cmd('tar xvf "{}" -C "{}"'.format(fpath2, dpath), verbout=1)

            # VOC 2007 test data
            fpath3 = ub.grabdata('http://host.robots.ox.ac.uk/pascal/VOC/voc2007/VOCtest_06-Nov-2007.tar', dpath=dpath)
            if force or not exists(join(dpath, 'VOCdevkit', 'VOC2007', 'ImageSets', 'Main', 'bird_test.txt')):
                ub.cmd('tar xvf "{}" -C "{}"'.format(fpath3, dpath), verbout=1)

        if 2012 in years:
            # VOC 2012 train+validation data
            fpath4 = ub.grabdata('https://pjreddie.com/media/files/VOCtrainval_11-May-2012.tar', dpath=dpath)
            if force or not exists(join(dpath, 'VOCdevkit', 'VOC2012', 'ImageSets', 'Main', 'bird_trainval.txt')):
                ub.cmd('tar xvf "{}" -C "{}"'.format(fpath4, dpath), verbout=1)
        return devkit_dpath

    def __nice__(self):
        return '{}'.format(len(self))

    def __len__(self):
        return len(self.gpaths)

    def __getitem__(self, index):
        """
        Returns:
            image, (bbox, class_idxs)

            bbox and class_idxs are variable-length
            bbox is in x1,y1,x2,y2 (i.e. tlbr) format

        CommandLine:
            xdoctest ~/code/netharn/netharn/data/voc.py VOCDataset.__getitem__ --show --voc

        Example:
            >>> # xdoc: +REQUIRES(--voc)
            >>> self = VOCDataset()
            >>> chw, label = self[1]
            >>> hwc = chw.numpy().transpose(1, 2, 0)
            >>> boxes, class_idxs = label
            >>> # xdoc: +REQUIRES(--show)
            >>> import netharn as nh
            >>> nh.util.qtensure()  # xdoc: +SKIP
            >>> nh.util.figure(fnum=1, doclf=True)
            >>> nh.util.imshow(hwc, colorspace='rgb')
            >>> nh.util.draw_boxes(boxes.numpy(), box_format='tlbr')
            >>> nh.util.show_if_requested()
        """
        if isinstance(index, tuple):
            # Get size index from the batch loader
            index, inp_size = index
        else:
            inp_size = self.base_wh
        hwc, boxes, gt_classes = self._load_item(index, inp_size)

        chw = torch.FloatTensor(hwc.transpose(2, 0, 1))
        gt_classes = torch.LongTensor(gt_classes)
        boxes = torch.FloatTensor(boxes)
        label = (boxes, gt_classes,)
        return chw, label

    def _load_item(self, index, inp_size=None):
        # from netharn.models.yolo2.utils.yolo import _offset_boxes
        import cv2
        image = self._load_image(index)
        annot = self._load_annotation(index)

        boxes = annot['boxes'].astype(np.float32)
        gt_classes = annot['gt_classes']

        # squish the bounding box and image into a standard size
        if inp_size is None:
            return image, boxes, gt_classes
        else:
            w, h = inp_size
            sx = float(w) / image.shape[1]
            sy = float(h) / image.shape[0]
            boxes[:, 0::2] *= sx
            boxes[:, 1::2] *= sy
            interpolation = cv2.INTER_AREA if (sx + sy) <= 2 else cv2.INTER_CUBIC
            hwc = cv2.resize(image, (w, h), interpolation=interpolation)
            return hwc, boxes, gt_classes

    def _load_image(self, index):
        import cv2
        fpath = self.gpaths[index]
        imbgr = cv2.imread(fpath, flags=cv2.IMREAD_COLOR)
        imrgb_255 = cv2.cvtColor(imbgr, cv2.COLOR_BGR2RGB)
        return imrgb_255

    def _load_annotation(self, index):
        import scipy
        import scipy.sparse
        import xml.etree.ElementTree as ET
        fpath = self.apaths[index]
        tree = ET.parse(fpath)
        objs = tree.findall('object')

        num_objs = len(objs)

        boxes = np.zeros((num_objs, 4), dtype=np.uint16)
        gt_classes = np.zeros((num_objs), dtype=np.int32)
        overlaps = np.zeros((num_objs, self.num_classes), dtype=np.float32)
        # "Seg" area for pascal is just the box area
        seg_areas = np.zeros((num_objs), dtype=np.float32)
        ishards = np.zeros((num_objs), dtype=np.int32)

        # Load object bounding boxes into a data frame.
        for ix, obj in enumerate(objs):
            bbox = obj.find('bndbox')
            # Make pixel indexes 0-based
            x1 = float(bbox.find('xmin').text) - 1
            y1 = float(bbox.find('ymin').text) - 1
            x2 = float(bbox.find('xmax').text) - 1
            y2 = float(bbox.find('ymax').text) - 1

            diffc = obj.find('difficult')
            difficult = 0 if diffc is None else int(diffc.text)
            ishards[ix] = difficult

            clsname = obj.find('name').text.lower().strip()
            cls = self._class_to_ind[clsname]
            boxes[ix, :] = [x1, y1, x2, y2]
            gt_classes[ix] = cls
            overlaps[ix, cls] = 1.0
            seg_areas[ix] = (x2 - x1 + 1) * (y2 - y1 + 1)

        overlaps = scipy.sparse.csr_matrix(overlaps)

        annots = {'boxes': boxes,
                  'gt_classes': gt_classes,
                  'gt_ishard': ishards,
                  'gt_overlaps': overlaps,
                  'flipped': False,
                  'fpath': fpath,
                  'seg_areas': seg_areas}
        return annots

    def show_image(self, index, fnum=None):
        from netharn import util
        hwc, boxes, gt_classes = self._load_item(index, inp_size=None)

        labels = list(ub.take(self.label_names, gt_classes))

        util.figure(doclf=True, fnum=fnum)
        util.imshow(hwc, colorspace='rgb')
        util.draw_boxes(boxes, color='green', box_format='tlbr', labels=labels)

    def make_loader(self, *args, **kwargs):
        """
        We need to do special collation to deal with different numbers of
        bboxes per item.

        Args:
            batch_size (int, optional): how many samples per batch to load
                (default: 1).
            shuffle (bool, optional): set to ``True`` to have the data
                reshuffled at every epoch (default: False).
            sampler (Sampler, optional): defines the strategy to draw samples
                from the dataset. If specified, ``shuffle`` must be False.
            batch_sampler (Sampler, optional): like sampler, but returns a
                batch of indices at a time. Mutually exclusive with batch_size,
                shuffle, sampler, and drop_last.
            num_workers (int, optional): how many subprocesses to use for data
                loading. 0 means that the data will be loaded in the main
                process.  (default: 0)
            pin_memory (bool, optional): If ``True``, the data loader will copy
                tensors into CUDA pinned memory before returning them.
            drop_last (bool, optional): set to ``True`` to drop the last
                incomplete batch, if the dataset size is not divisible by the
                batch size. If ``False`` and the size of dataset is not
                divisible by the batch size, then the last batch will be
                smaller. (default: False)
            timeout (numeric, optional): if positive, the timeout value for
                collecting a batch from workers. Should always be non-negative.
                (default: 0)
            worker_init_fn (callable, optional): If not None, this will be
                called on each worker subprocess with the worker id (an int in
                ``[0, num_workers - 1]``) as input, after seeding and before
                data loading. (default: None)

        References:
            https://github.com/pytorch/pytorch/issues/1512

        Example:
            >>> # xdoc: +REQUIRES(--voc)
            >>> self = VOCDataset()
            >>> #inbatch = [self[i] for i in range(10)]
            >>> loader = self.make_loader(batch_size=10)
            >>> batch = next(iter(loader))
            >>> images, labels = batch
            >>> assert len(images) == 10
            >>> assert len(labels) == 2
            >>> assert len(labels[0]) == len(images)
        """
        kwargs['collate_fn'] = collate.list_collate
        loader = torch_data.DataLoader(self, *args, **kwargs)
        return loader

    def to_coco(self):
        """ Transform VOC to coco-style dataset """
        import netharn as nh
        voc_dset = self
        coco_dset = nh.data.coco_api.CocoDataset()
        coco_dset._build_index()

        for cx, catname in enumerate(self.label_names):
            coco_dset.add_category(catname, cid=int(cx))

        for gx, gpath in enumerate(ub.ProgIter(voc_dset.gpaths,
                                               label='convert coco')):
            coco_dset.add_image(gpath, gid=int(gx))
            voc_anno = self._load_annotation(gx)

            for i in range(len(voc_anno['boxes'])):
                box = nh.util.Boxes(voc_anno['boxes'][i], 'tlbr')
                cx = voc_anno['gt_classes'][i]
                weight = 1 - voc_anno['gt_ishard'][i]
                coco_dset.add_annotation(
                    gid=int(gx), cid=int(cx), bbox=box, weight=weight)
        return coco_dset


if __name__ == '__main__':
    r"""
    CommandLine:
        python -m netharn.data.voc all
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



# -*- coding: utf-8 -*-
"""
mkinit netharn.initializers
"""
from __future__ import absolute_import, division, print_function, unicode_literals

from netharn.api import Initializer  # NOQA

# backwards compatibility patch to support older deployed models
from netharn.initializers import core
nninit_base = core
nninit_base._BaseInitializer = Initializer

__explicit__ = ['Initializer', 'nninit_base']


# <AUTOGEN_INIT>
from netharn.initializers import core
from netharn.initializers import functional
from netharn.initializers import lsuv
from netharn.initializers import pretrained

from netharn.initializers.core import (KaimingNormal, KaimingUniform, NoOp,
                                       Orthogonal,)
from netharn.initializers.functional import (apply_initializer,
                                             load_partial_state,
                                             trainable_layers,)
from netharn.initializers.lsuv import (LSUV, Orthonormal, svd_orthonormal,)
from netharn.initializers.pretrained import (Pretrained,)

__all__ = ['Initializer', 'KaimingNormal', 'KaimingUniform', 'LSUV', 'NoOp',
           'Orthogonal', 'Orthonormal', 'Pretrained', 'apply_initializer',
           'core', 'functional', 'load_partial_state', 'lsuv', 'pretrained',
           'svd_orthonormal', 'trainable_layers']
# </AUTOGEN_INIT>



import numpy as np
import warnings
from scipy.sparse import coo_matrix


def fast_confusion_matrix(y_true, y_pred, n_labels, sample_weight=None):
    """
    faster version of sklearn confusion matrix that avoids the
    expensive checks and label rectification

    Args:
        y_true (ndarray[int]): ground truth class label for each sample
        y_pred (ndarray[int]): predicted class label for each sample
        n_labels (int): number of labels
        sample_weight (ndarray[int|float]): weight of each sample

    Returns:
        ndarray[int64|float64, dim=2]:
            matrix where rows represent real and cols represent pred and the
            value at each cell is the total amount of weight

    Example:
        >>> y_true = np.array([0, 0, 0, 0, 1, 1, 1, 0,  0, 1])
        >>> y_pred = np.array([0, 0, 0, 0, 0, 0, 0, 1,  1, 1])
        >>> fast_confusion_matrix(y_true, y_pred, 2)
        array([[4, 2],
               [3, 1]])
        >>> fast_confusion_matrix(y_true, y_pred, 2).ravel()
        array([4, 2, 3, 1])
    """
    if sample_weight is None:
        sample_weight = np.ones(len(y_true), dtype=np.uint8)
    # The accumulation dtype needs to have 64bits to avoid overflow
    dtype = np.float64 if sample_weight.dtype.kind == 'f' else np.int64
    matrix = coo_matrix((sample_weight, (y_true, y_pred)),
                        shape=(n_labels, n_labels),
                        dtype=dtype).toarray()
    return matrix


def _truncated_roc(y_df, bg_idx=-1, fp_cutoff=None):
    """
    Computes truncated ROC info
    """
    import sklearn
    try:
        from sklearn.metrics._ranking import _binary_clf_curve
    except ImportError:
        from sklearn.metrics.ranking import _binary_clf_curve
    y_true = (y_df['true'] == y_df['pred'])
    y_score = y_df['score']
    sample_weight = y_df['weight']

    # y_true[y_true == -1] = 0

    # < TRUCNATED PART >
    # GET ROC CURVES AT A PARTICULAR FALSE POSITIVE COUNT CUTOFF
    # This will let different runs be more comparable
    realpos_total = sample_weight[(y_df['txs'] >= 0)].sum()

    fp_count, tp_count, count_thresholds = _binary_clf_curve(
        y_true, y_score, pos_label=1, sample_weight=sample_weight)

    if len(count_thresholds) > 0 and count_thresholds[-1] == 0:
        # Chop off the last entry where it will jump
        count_thresholds = count_thresholds[:-1]
        tp_count = tp_count[:-1]
        fp_count = fp_count[:-1]

    # Cutoff the curves at a comparable point
    if fp_cutoff is None:
        fp_cutoff = np.inf
    idxs = np.where(fp_count > fp_cutoff)[0]
    if len(idxs) == 0:
        idx = len(fp_count)
    else:
        idx = idxs[0]
    trunc_fp_count = fp_count[:idx]
    trunc_tp_count = tp_count[:idx]
    trunc_thresholds = count_thresholds[:idx]

    # if the cuttoff was not reached, horizontally extend the curve
    # This will hurt the scores (aka we may be bias against small
    # scenes), but this will ensure that big scenes are comparable
    if len(fp_count) == 0:
        trunc_fp_count = np.array([fp_cutoff])
        trunc_tp_count = np.array([0])
        trunc_thresholds = np.array([0])
        # THIS WILL CAUSE AUC TO RAISE AN ERROR IF IT GETS HIT
    elif fp_count[-1] < fp_cutoff and np.isfinite(fp_cutoff):
        trunc_fp_count = np.hstack([trunc_fp_count, [fp_cutoff]])
        trunc_tp_count = np.hstack([trunc_tp_count, [trunc_tp_count[-1]]])
        trunc_thresholds = np.hstack([trunc_thresholds, [0]])

    falsepos_total = trunc_fp_count[-1]  # is this right?

    trunc_tpr = trunc_tp_count / realpos_total
    trunc_fpr = trunc_fp_count / falsepos_total
    trunc_auc = sklearn.metrics.auc(trunc_fpr, trunc_tpr)
    # < /TRUCNATED PART >
    roc_info = {
        'fp_cutoff': fp_cutoff,
        'realpos_total': realpos_total,
        'tpr': trunc_tpr,
        'fpr': trunc_fpr,
        'fp_count': trunc_fp_count,
        'tp_count': trunc_tp_count,
        'thresholds': trunc_thresholds,
        'auc': trunc_auc,
    }
    return roc_info


def _pr_curves(y):
    """
    Compute a PR curve from a method

    Args:
        y (pd.DataFrame | DataFrameArray): output of detection_confusions

    Returns:
        Tuple[float, ndarray, ndarray]

    Example:
        >>> # xdoctest: +REQUIRES(module:sklearn)
        >>> import pandas as pd
        >>> y1 = pd.DataFrame.from_records([
        >>>     {'pred': 0, 'score': 10.00, 'true': -1, 'weight': 1.00},
        >>>     {'pred': 0, 'score':  1.65, 'true':  0, 'weight': 1.00},
        >>>     {'pred': 0, 'score':  8.64, 'true': -1, 'weight': 1.00},
        >>>     {'pred': 0, 'score':  3.97, 'true':  0, 'weight': 1.00},
        >>>     {'pred': 0, 'score':  1.68, 'true':  0, 'weight': 1.00},
        >>>     {'pred': 0, 'score':  5.06, 'true':  0, 'weight': 1.00},
        >>>     {'pred': 0, 'score':  0.25, 'true':  0, 'weight': 1.00},
        >>>     {'pred': 0, 'score':  1.75, 'true':  0, 'weight': 1.00},
        >>>     {'pred': 0, 'score':  8.52, 'true':  0, 'weight': 1.00},
        >>>     {'pred': 0, 'score':  5.20, 'true':  0, 'weight': 1.00},
        >>> ])
        >>> import netharn as nh
        >>> import kwarray
        >>> y2 = kwarray.DataFrameArray(y1)
        >>> _pr_curves(y2)
        >>> _pr_curves(y1)
    """
    import sklearn
    # compute metrics on a per class basis
    if y is None:
        return np.nan, [], []

    # References [Manning2008] and [Everingham2010] present alternative
    # variants of AP that interpolate the precision-recall curve. Currently,
    # average_precision_score does not implement any interpolated variant
    # http://scikit-learn.org/stable/modules/model_evaluation.html

    # In the future, we should simply use the sklearn version
    # which gives nice easy to reproduce results.
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', message='invalid .* true_divide')
        is_correct = (y['true'] == y['pred']).astype(np.int)
        ap = sklearn.metrics.average_precision_score(
            y_true=is_correct, y_score=y['score'],
            sample_weight=y['weight'],
        )
        prec, rec, thresholds = sklearn.metrics.precision_recall_curve(
            is_correct, y['score'], sample_weight=y['weight'],
        )
    return ap, prec, rec


def _average_precision(tpr, ppv):
    """
    Compute average precision of a binary PR curve. This is simply the area
    under the curve.

    Args:
        tpr (ndarray): true positive rate - aka recall
        ppv (ndarray): positive predictive value - aka precision
    """
    # The average precision is simply the area under the PR curve.
    xdata = tpr
    ydata = ppv
    if xdata[0] > xdata[-1]:
        xdata = xdata[::-1]
        ydata = ydata[::-1]
    # Note: we could simply use sklearn.metrics.auc, which has more robust
    # checks.
    ap = np.trapz(y=ydata, x=xdata)
    return ap



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
import ubelt as ub
import numpy as np


class _Scheduler(object):
    """
    New Base object for scheduling optimizer parameters
    """

    __netharn_redesign__ = True

    def __init__(self, optimizer=None):
        self.optimizer = optimizer
        self.epoch = 0
        if self.optimizer is not None:
            self.step(self.epoch)

    def _optim_attrs(self):
        """ returns a collated dictionary of non-params optimizer attributes """
        keys = self.optimizer.param_groups[0].keys()
        return {k: _get_optimizer_values(self.optimizer, k)
                for k in keys if k != 'params'}

    def state_dict(self):
        """Returns the state of the scheduler as a :class:`dict`.

        It contains an entry for every variable in self.__dict__ which
        is not the optimizer.
        """
        return {k: v for k, v in self.__dict__.items() if k != 'optimizer'}

    def load_state_dict(self, state_dict):
        """Loads the schedulers state.

        Arguments:
            state_dict (dict): scheduler state. Should be an object returned
                from a call to :meth:`state_dict`.
        """
        self.__dict__.update(state_dict)

    def step(self, epoch=None):
        """
        Simply updates the attributes (lr, momentum, etc.) of the optimizer to
        the current epoch.
        """
        raise NotImplementedError


class ListedScheduler(_Scheduler):
    """
    List key transitions points for every optimizer attribute.

    Notes:
        attributes for SGD are:
            'lr'
            'momentum'
            'dampening'
            'weight_decay'
            'nesterov'

    Example:
        >>> from netharn.schedulers.scheduler_redesign import *
        >>> import ubelt as ub
        >>> import netharn as nh
        >>> import torch
        >>> import copy
        >>> model = nh.models.ToyNet2d()
        >>> optimizer = torch.optim.SGD(model.parameters(), lr=0)
        >>> points = {
        >>>     'lr'       : {0: 1, 2: 21, 3: 31, 6: 61, 9: 91},
        >>>     'momentum' : {0: 2, 2: 22, 3: 32, 6: 62, 9: 92},
        >>> }
        >>> # Test SGD
        >>> self = ListedScheduler(points, optimizer=optimizer, interpolation='left')
        >>> states = []
        >>> for _ in range(4):
        ...     states.append((self.epoch, copy.deepcopy(self._optim_attrs())))
        ...     self.step()
        >>> print('states = {}'.format(ub.repr2(states, nl=1)))
        states = [
            (0, {'dampening': [0], 'lr': [1], 'momentum': [2], 'nesterov': [False], 'weight_decay': [0]}),
            (1, {'dampening': [0], 'lr': [1], 'momentum': [2], 'nesterov': [False], 'weight_decay': [0]}),
            (2, {'dampening': [0], 'lr': [21], 'momentum': [22], 'nesterov': [False], 'weight_decay': [0]}),
            (3, {'dampening': [0], 'lr': [31], 'momentum': [32], 'nesterov': [False], 'weight_decay': [0]}),
        ]
        >>> # Test ADAM
        >>> optimizer = torch.optim.Adam(model.parameters(), lr=0)
        >>> self = ListedScheduler(points, optimizer=optimizer, interpolation='linear')
        >>> states = []
        >>> for _ in range(4):
        ...     states.append((self.epoch, copy.deepcopy(self._optim_attrs())))
        ...     self.step()
        >>> print('states = {}'.format(ub.repr2(states, nl=1)))
        states = [
            (0, {'amsgrad': [False], 'betas': [(2, 0.999)], 'eps': [1e-08], 'lr': [1], 'weight_decay': [0]}),
            (1, {'amsgrad': [False], 'betas': [(12.0, 0.999)], 'eps': [1e-08], 'lr': [11.0], 'weight_decay': [0]}),
            (2, {'amsgrad': [False], 'betas': [(22, 0.999)], 'eps': [1e-08], 'lr': [21], 'weight_decay': [0]}),
            (3, {'amsgrad': [False], 'betas': [(32, 0.999)], 'eps': [1e-08], 'lr': [31], 'weight_decay': [0]}),
        ]
    """
    def __init__(self, points, interpolation='linear', optimizer=None):
        """
        Args:
            points (Dict[str, Dict[int, float]):
                dictionary mapping optimizer attribute names (e.g. lr) to a
                dictinoary mapping epochs to values.

            interpolation (str, default=linear):
                one of: left, nearest, linear,
        """
        self.points = points
        if not interpolation:
            interpolation = 'left'
        self.interpolation = interpolation
        super(ListedScheduler, self).__init__(optimizer=optimizer)

    def get_value(self, attr='lr', epoch=None):
        if epoch is None:
            epoch = self.epoch + 1
        points = self.points[attr]
        value = _interpolate(points, epoch, interpolation=self.interpolation)
        # print('attr, epoch, value = {}, {}, {}'.format(attr, epoch, value))
        return value

    def step(self, epoch=None):
        if epoch is None:
            epoch = self.epoch + 1
        self.epoch = epoch
        for attr in self.points.keys():
            values = self.get_value(attr, epoch=epoch)
            _set_optimizer_values(self.optimizer, attr, values)


def _set_optimizer_values(optimizer, attr, value_or_values):
    """ sets the optimizer value from each param group """
    if ub.iterable(value_or_values):
        values = value_or_values
    else:
        values = [value_or_values] * len(optimizer.param_groups)
    for group, value in zip(optimizer.param_groups, values):
        if attr in group:
            group[attr] = value
        else:
            # Allow alias attr names to be used similar to fastai
            if attr == 'momentum' and 'betas' in group:
                momentum, beta = group['betas']
                group['betas'] = (value, beta)
            elif attr == 'beta' and 'betas' in group:
                momentum, beta = group['betas']
                group['betas'] = (momentum, value)
            else:
                raise KeyError(attr)


def _get_optimizer_values(optimizer, attr):
    """ gets the optimizer value from each param group """
    try:
        return [group[attr] for group in optimizer.param_groups]
    except KeyError:
        group = optimizer.param_groups[0]
        if attr == 'momentum' and 'betas' in group:
            return [group['betas'][0] for group in optimizer.param_groups]
        if attr == 'beta' and 'betas' in group:
            return [group['betas'][1] for group in optimizer.param_groups]


def _interpolate(points, x, interpolation='linear'):
    """
    Interpolates between sparse keypoints

    Args:
        points (Dict[float, float]): Mapping from x-axis to y-axis
        x (float): point on the x-axis to get a value for
        interpolation (str): how to interpolate between keypoints

    Returns:
        float: the interpolated value

    Example:
        >>> points = {0: .01, 2: .02, 3: .1, 6: .05, 9: .025}
        >>> _interpolate(points, x=1, interpolation='left')
        0.01
        >>> _interpolate(points, x=1, interpolation='linear')
        0.015
        >>> _interpolate(points, x=1.9, interpolation='nearest')
        0.02
    """
    keys = sorted(points.keys())

    if x in keys:
        key_left = x
        key_right = x
    else:
        idx = np.searchsorted(keys, x, 'left') - 1
        key_left = keys[idx]
        if idx < len(keys) - 1:
            key_right = keys[idx + 1]
        else:
            key_right = key_left
        if x < key_left:
            key_left = min(points.keys())

    if key_left == key_right:
        value = points[key_left]
    else:
        # Interpolate between the keypoints to sample an sub-point value
        if interpolation == 'left':
            value = points[key_left]
        elif interpolation == 'nearest':
            if (x - key_left) <= (key_right - x):
                value = points[key_left]
            else:
                value = points[key_right]
        elif interpolation == 'linear':
            value_left = points[key_left]
            value_right = points[key_right]
            alpha = (x - key_left) / float((key_right - key_left))
            value = alpha * value_right + (1 - alpha) * value_left
        else:
            raise KeyError(interpolation)
        # elif interpolation.startswith('log'):
        #     value_left = points[key_left]
        #     value_right = points[key_right]
        #     log = {
        #         'log': np.log,
        #         'log2': np.log2,
        #         'log10': np.log10,
        #     }[interpolation]
        #     if value_left < value_right:
        #         np.logspace(log(value_left), log(value_right))
        #     else:
        #         np.logspace(log(value_right), log(value_left))
    return value



"""
mkinit netharn.data.transforms')"
"""
# flake8: noqa

__DYNAMIC__ = False
if __DYNAMIC__:
    from mkinit import dynamic_init
    exec(dynamic_init(__name__))
else:
    # <AUTOGEN_INIT>
    from netharn.data.transforms import augmenter_base
    from netharn.data.transforms import augmenters
    from netharn.data.transforms.augmenter_base import (ParamatarizedAugmenter,)
    from netharn.data.transforms.augmenters import (HSVShift, LetterboxResize,
                                                    Resize, demodata_hsv_image,)
    # </AUTOGEN_INIT>



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
import ubelt as ub
import math
import torch
import torch.nn.functional as F
import torch.nn as nn
import torchvision
from collections import OrderedDict
from six.moves import builtins
from netharn.analytic import analytic_for
# try:
from netharn.device import DataSerial
# except ImportError:
#     DataSerial = None

REGISTERED_TYPES = []


SHAPE_CLS = tuple  # We exepct shapes to be specified as this class


def compute_type(*types):
    def _wrap(func):
        for type in types:
            if type is not None:
                REGISTERED_TYPES.append((type, func))
        return func
    return _wrap


def output_shape_of(outputs):
    """
    Given a network output, try and find the shape. Works in most standard
    cases, but not all cases.

    Args:
        outputs (Tensor | Dict | Tuple): some typical torch network output

    Example:
        >>> output_shape_of(torch.empty(3, 2))
        (3, 2)
        >>> output_shape_of({'a': torch.empty(3, 2)})
        {'a': (3, 2)}
        >>> output_shape_of(((torch.empty(3, 2),),))
        [[(3, 2)]]
    """
    if torch.is_tensor(outputs):
        computed_output_shape = SHAPE_CLS(outputs.shape)
    elif isinstance(outputs, dict):
        dict_cls = outputs.__class__  # handle odict
        computed_output_shape = dict_cls([
            (k, output_shape_of(v)) for k, v in outputs.items()])
    elif isinstance(outputs, tuple):
        # Allow outputs to be a tuple of tensors
        computed_output_shape = [output_shape_of(o) for o in outputs]
    else:
        raise TypeError('Cannot find shape of {!r}'.format(type(outputs)))
    return computed_output_shape


def _brute_force_output_shape_for(self, input_shape):
    """
    Computes output shape by actually running the network. Works in most
    standard cases, but not all cases. If the batch size is None, we attempt to
    be smart about ensuring that that None is propogated in the output.

    Example:
        >>> module = nn.Conv2d(3, 11, 3, 1, 0)
        >>> _brute_force_output_shape_for(module, (None, 3, 256, 256))
        (None, 11, 254, 254)
    """
    _input_shape = list(input_shape)
    unknown_bsize = _input_shape[0] is None
    if unknown_bsize:
        bsize = 2
        _input_shape[0] = bsize
    device = next(iter(self.state_dict().values())).device
    dummy_input = torch.rand(*_input_shape).to(device)
    dummy_output = self(dummy_input)
    output_shape = output_shape_of(dummy_output)
    if torch.is_tensor(dummy_output):
        if unknown_bsize:
            if output_shape[0] == bsize:
                output_shape = list(output_shape)
                output_shape[0] = None
        output_shape = SHAPE_CLS(output_shape)
    else:
        raise NotImplementedError('other output types')
    return output_shape


def _simplify(shape):
    import sympy
    if isinstance(shape, (tuple, list)):
        shape = shape.__class__([_simplify(v) for v in shape])
    elif isinstance(shape, dict):
        shape = shape.__class__([(k, _simplify(v)) for k, v in shape.items()])
    elif isinstance(shape, sympy.Expr):
        shape = sympy.simplify(shape)
    return shape


class HiddenShapes(analytic_for.Hidden):
    """
    Augments normal hidden shape dicts with a convinience setitem

    Doctest:
        >>> from netharn.analytic.output_shape_for import *
        >>> shape = OutputShape.coerce([None, 3, 32, 32], 'foo')
        >>> print(HiddenShapes({'e': shape}))
        <HiddenShapes({'e': 'foo'})>
        >>> hidden = HiddenShapes({'a': 1})
        >>> hidden['b'] = 2
        >>> hidden['c'] = shape
        >>> print(hidden)
        <HiddenShapes({'a': 1, 'b': 2, 'c': 'foo'})>
    """
    pass


# class HiddenShapes(OrderedDict, ub.NiceRepr):
#     """
#     Augments normal hidden shape dicts with a convinience setitem

#     Doctest:
#         >>> from netharn.analytic.output_shape_for import *
#         >>> shape = OutputShape.coerce([None, 3, 32, 32], 'foo')
#         >>> print(HiddenShapes({'e': shape}))
#         <HiddenShapes({'e': 'foo'})>
#         >>> hidden = HiddenShapes({'a': 1})
#         >>> hidden['b'] = 2
#         >>> hidden['c'] = shape
#         >>> print(hidden)
#         <HiddenShapes({'a': 1, 'b': 2, 'c': 'foo'})>
#     """
#     def __nice__(self):
#         return ub.repr2(self, nl=0)

#     def __str__(self):
#         return ub.NiceRepr.__str__(self)

#     def __repr__(self):
#         return ub.NiceRepr.__repr__(self)

#     def __setitem__(self, key, value):
#         if getattr(value, 'hidden', None) is not None:
#             # When setting a value to an OutputShape object, if that object has
#             # a hidden shape, then use that instead.
#             value = value.hidden
#         return OrderedDict.__setitem__(self, key, value)

#     def shallow(self, n=1):
#         """
#         Grabs only the shallowest n layers of hidden shapes
#         """
#         if n == 0:
#             last = self
#             while isinstance(last, HiddenShapes):
#                 values = list(last.values())
#                 if len(values):
#                     last = values[-1]
#                 else:
#                     break
#             return last
#         else:
#             output = OrderedDict()
#             for key, value in self.items():
#                 # if isinstance(value, HiddenShapes):
#                 if hasattr(value, 'shallow'):
#                     value = value.shallow(n - 1)
#                 output[key] = value
#             return output


class OutputShape(analytic_for.Output):
    """
    Mixin class to extend output shapes with extra information

    Doctest:
        >>> from netharn.analytic.output_shape_for import *
        >>> shape = OutputShape.coerce([None, 3, 32, 32], 'foo')
        >>> print('shape = {!r}'.format(shape))
        shape = (None, 3, 32, 32)
        >>> print('shape.hidden = {!r}'.format(shape.hidden))
        shape.hidden = 'foo'
    """
    def __init__(self, data=None, hidden=None):
        self.data = data
        self.hidden = hidden

    @classmethod
    def template(cls, type):
        """ Get a specific template for a subclass type """
        if type is tuple:
            return OutputShapeTuple
        elif type is OrderedDict:
            return OutputShapeDict
        elif type is dict:
            return OutputShapeDict
        else:
            raise TypeError(type)

    @classmethod
    def coerce(cls, data=None, hidden=None):
        """
        Create an OutputShape instance of the approriate subclass given the
        type of input data.
        """
        if isinstance(data, cls):
            if hidden is None:
                self = data
            else:
                self = data.__class__(data, hidden)
        elif isinstance(data, (tuple, list)):
            self = cls.template(tuple)(data, hidden)
        elif isinstance(data, dict):
            self = cls.template(dict)(data, hidden)
        else:
            raise TypeError(type(data))
        return self


class OutputShapeTuple(tuple, OutputShape):
    """ OutputShape templated as a tuple """
    def __new__(cls, data=None, hidden=None):
        # tuple subclass is a bit weird
        if data is None:
            data = tuple()
        self = tuple.__new__(OutputShapeTuple, data)
        OutputShape.__init__(self, data, hidden)
        return self


class OutputShapeDict(OrderedDict, OutputShape):
    """ OutputShape templated as a dictionary """
    def __init__(self, data=None, hidden=None):
        if data is None:
            data = OrderedDict()
        OrderedDict.__init__(self, data)
        OutputShape.__init__(self, data, hidden)


class OutputShapeFor(analytic_for.OutputFor):
    """
    Compute the output shape for standard torch modules as well as
    any custom modules that follow the OutputShapeFor protocol.

    Notes:
        The OutputShapeFor protocol is simple. For any custom torch module
        define the method `output_shape_for(self, input_shape)`, which is
        typically written to mirror the `forward` function. Instead of calling
        forward on the custom module's torch members use `OutputShapeFor`. See
        netharn.layers for more examples of custom layers that implement this
        protocol. A simple example is shown below.

    Example:
        >>> # Example showing how to implement the OutputShapeFor protocol
        >>> class MyCustomNet(nn.Module):
        >>>     def __init__(self):
        >>>         super(MyCustomNet, self).__init__()
        >>>         self.conv1 = nn.Conv2d(1, 5, 3)
        >>>         self.pool1 = nn.MaxPool2d(2)
        >>>         self.conv2 = nn.Conv2d(5, 7, 3)
        >>>     def forward(self, input):
        >>>         x = input
        >>>         x = self.conv1(x)
        >>>         x = self.pool1(x)
        >>>         x = self.conv2(x)
        >>>         return x
        >>>     def output_shape_for(self, input_shape):
        >>>         x = input_shape
        >>>         # Note using hidden shapes is optional, but sometimes useful
        >>>         hidden = HiddenShapes()
        >>>         # The basic idea is to simply mirror the forward func
        >>>         # but instead of calling the modules use output shape for
        >>>         hidden['conv1'] = x = OutputShapeFor(self.conv1)(x)
        >>>         hidden['pool1'] = x = OutputShapeFor(self.pool1)(x)
        >>>         hidden['conv2'] = x = OutputShapeFor(self.conv2)(x)
        >>>         shape = OutputShape.coerce(x, hidden)
        >>>         return shape
        >>> net = MyCustomNet()
        >>> # Now it is very easy and efficient to infer the output shape
        >>> input_shape = (None, 1, 9, 9)
        >>> net.output_shape_for(input_shape)
        (None, 7, 1, 1)
        >>> # The OutputShapeFor class now recognizes your module as well
        >>> # so it can be used to constuct more complex modules while
        >>> # still maintaining the ability fo infer the output shape.
        >>> OutputShapeFor(net)(input_shape)
        (None, 7, 1, 1)
        >>> # Note that if you did return an true OutputShape object with
        >>> # a populated hidden shape attribute, then you can access it
        >>> # to inspect how the shape changes in the hidden layer of the net
        >>> print(OutputShapeFor(net)(input_shape).hidden)
        <HiddenShapes({'conv1': (None, 5, 7, 7), 'pool1': (None, 5, 3, 3), 'conv2': (None, 7, 1, 1)})>

    Example:
        >>> # Example showing how this class is used on basic torch Modules
        >>> module = nn.Conv2d(3, 11, 3, 1, 0)
        >>> OutputShapeFor(module)((1, 3, 256, 256))
        (1, 11, 254, 254)
    """
    math = math  # for hacking in sympy

    def __init__(self, module, force=False):
        """
        Args:
            module (nn.Module) : module with output_shape_for func or
                with some known registered type (e.g. torch.nn.Conv2d).

            force (bool): if True and no implicit computation is known
                try to create a dummy input with input_shape and simply
                run it through the network to see what shape it produces.
                (Defaults to False).
        """
        self._requires_force = False
        self.module = module
        # First try to lookup the output_shape_for func
        self._func = getattr(module, 'output_shape_for', None)

        if self._func is None:
            # Lookup shape func if we can't find it
            found = []
            for type, _func in REGISTERED_TYPES:
                try:
                    if module is type or isinstance(module, type):
                        found.append(_func)
                except TypeError:
                    pass
            if len(set(found)) == 1:
                self._func = found[0]
            elif len(found) == 0:
                raise TypeError('Unknown (output_shape) module type {}'.format(module))
            else:
                raise AssertionError('Ambiguous (output_shape) module {}. Found {}'.format(module, found))

    def __call__(self, *args, **kwargs):
        if isinstance(self.module, nn.Module):
            # bound methods dont need module
            is_bound  = hasattr(self._func, '__func__') and getattr(self._func, '__func__', None) is not None
            is_bound |= hasattr(self._func, 'im_func') and getattr(self._func, 'im_func', None) is not None
            if is_bound:
                output_shape = self._func(*args, **kwargs)
            else:
                # nn.Module with state
                output_shape = self._func(self.module, *args, **kwargs)
        else:
            # a simple pytorch func
            output_shape = self._func(*args, **kwargs)

        # Package the output shape up in the appropriate wrapper class
        output_shape = OutputShape.coerce(output_shape)
        # if self.math.__name__ == 'sympy':
        #     output_shape = _simplify(output_shape)
        # debug = True
        # if debug:
        #     print('{}.output_shape = {}'.format(str(self._func.__name__), output_shape))
        return output_shape

    def _check_consistency(self, input_shape, **kwargs):
        """
        Test function to check that expected shape is equal to computed shape.
        The kwargs are passed to both output_shape_for and forward, so ensure
        that both functions accept the same arguments.
        """
        # Run the output shape computation
        expected = self(input_shape, **kwargs)

        if isinstance(expected, OutputShape):
            expected_output_shape = expected.data
        else:
            expected_output_shape = expected

        # Create dummy inputs and send them through the network
        inputs = torch.randn(input_shape)
        with torch.no_grad():
            self.module.eval()
            outputs = self.module(inputs, **kwargs)

        if isinstance(outputs, dict):
            if not isinstance(expected_output_shape, dict):
                raise AssertionError((
                    'if outputs is a dict, then output_shape must also be '
                    'a corresponding dict. Instead we got: '
                    'type(outputs)={} '
                    'type(expected_output_shape)={} '
                ).format(type(outputs), type(expected_output_shape)))
        computed_output_shape = output_shape_of(outputs)

        if computed_output_shape != expected_output_shape:
            print('expected_output_shape = {}'.format(ub.repr2(expected_output_shape, nl=0)))
            print('computed_output_shape = {}'.format(ub.repr2(computed_output_shape, nl=0)))
            raise AssertionError(
                'computed shape {!r} != expected shape {!r}'.format(
                    computed_output_shape,
                    expected_output_shape,
                )
            )
        return expected_output_shape

    @staticmethod
    @compute_type(nn.Upsample)
    def Upsample(module, input_shape):
        r"""
        - Input: :math:`(N, C, H_{in}, W_{in})`
        - Output: :math:`(N, C, H_{out}, W_{out})` where
            :math:`H_{out} = floor(H_{in} * scale\_factor)`
            :math:`W_{out} = floor(W_{in}  * scale\_factor)`

        Example:
            >>> # xdoctest: +SKIP
            >>> # There is a torch bug in 1.1.0 that prevents this from working
            >>> from netharn.analytic.output_shape_for import *
            >>> input_shape = (1, 3, 256, 256, 256)
            >>> module = nn.Upsample(scale_factor=(2, 3, 4))
            >>> output_shape = OutputShapeFor(module)(input_shape)
            >>> print('output_shape = {!r}'.format(output_shape))
            output_shape = (1, 3, 512, 768, 1024)
            >>> module = nn.Upsample(size=100)
            >>> output_shape = OutputShapeFor(module)(input_shape)
            >>> print('output_shape = {!r}'.format(output_shape))
            output_shape = (1, 3, 100, 100, 100)
            >>> input_shape = (1, 3, 256, 256)
            >>> module = nn.UpsamplingBilinear2d(scale_factor=2)
            >>> output_shape = OutputShapeFor(module)(input_shape)
            >>> print('output_shape = {!r}'.format(output_shape))
            output_shape = (1, 3, 512, 512)
        """
        math = OutputShapeFor.math
        # N, C, *DIMS_in = input_shape
        N, C = input_shape[0:2]
        DIMS_in = input_shape[2:]

        if module.size is None:
            scale_factor = ensure_iterablen(module.scale_factor, len(DIMS_in))
            int = builtins.int if math.__name__ == 'math' else ub.identity
            DIMS_out = [
                int(math.floor(D_in * scale_factor[i]))
                for i, D_in in enumerate(DIMS_in)
            ]
        else:
            DIMS_out = ensure_iterablen(module.size, len(DIMS_in))

        output_shape = SHAPE_CLS([N, C] + list(DIMS_out))
        if math.__name__ == 'sympy':
            output_shape = _simplify(output_shape)
        return output_shape

    @staticmethod
    @compute_type(torch.nn.functional.interpolate)
    def interpolate(input_shape, size=None, scale_factor=None, **kwargs):
        """
        Example:
            >>> from netharn.analytic.output_shape_for import *
            >>> input_shape = (1, 3, 256, 256)
            >>> output_shape = OutputShapeFor(torch.nn.functional.interpolate)(input_shape, size=(32, 32))
            >>> print('output_shape = {!r}'.format(output_shape))
            output_shape = (1, 3, 32, 32)
        """
        math = OutputShapeFor.math
        # N, C, *DIMS_in = input_shape
        N, C = input_shape[0:2]
        DIMS_in = input_shape[2:]

        if size is None:
            scale_factor = ensure_iterablen(scale_factor, len(DIMS_in))
            int = builtins.int if math.__name__ == 'math' else ub.identity
            DIMS_out = [
                int(math.floor(D_in * scale_factor[i]))
                for i, D_in in enumerate(DIMS_in)
            ]
        else:
            DIMS_out = ensure_iterablen(size, len(DIMS_in))

        output_shape = SHAPE_CLS([N, C] + list(DIMS_out))
        if math.__name__ == 'sympy':
            output_shape = _simplify(output_shape)
        return output_shape

    @staticmethod
    @compute_type(nn.ConvTranspose1d)
    def conv1dT(module, input_shape):
        return OutputShapeFor.convndT(module, input_shape, 1)

    @staticmethod
    @compute_type(nn.ConvTranspose2d)
    def conv2dT(module, input_shape):
        return OutputShapeFor.convndT(module, input_shape, 2)

    @staticmethod
    @compute_type(nn.ConvTranspose3d)
    def conv3dT(module, input_shape):
        return OutputShapeFor.convndT(module, input_shape, 3)

    @staticmethod
    @compute_type(nn.Conv1d)
    def conv1d(module, input_shape):
        return OutputShapeFor.convnd(module, input_shape, 1)

    @staticmethod
    @compute_type(nn.Conv2d)
    def conv2d(module, input_shape):
        return OutputShapeFor.convnd(module, input_shape, 2)

    @staticmethod
    @compute_type(nn.ZeroPad2d)
    def zeropad2d(module, input_shape):
        r"""
        Shape:
            - Input: :math:`(N, C, H_{in}, W_{in})`
            - Output: :math:`(N, C, H_{out}, W_{out})` where

              :math:`H_{out} = H_{in} + \text{padding\_top} + \text{padding\_bottom}`

              :math:`W_{out} = W_{in} + \text{padding\_left} + \text{padding\_right}`

        Example:
            >>> module = nn.ZeroPad2d([2, 3, 5, 7])
            >>> input_shape = (1, 3, 5, 7)
            >>> out = OutputShapeFor(module)(input_shape)
            >>> out_want = module(torch.zeros(*input_shape)).shape
            >>> assert out == tuple(out_want)
        """
        return OutputShapeFor.pad(input_shape, module.padding)

    @staticmethod
    @compute_type(F.conv2d)
    def f_conv2d(inputs, weight, bias=None, stride=1, padding=0,
                 dilation=1, groups=1):
        """
        Example:
            >>> x = inputs = (1, 124, 226, 226)
            >>> module = nn.Conv2d(128, 64, kernel_size=(3, 5), groups=8)
            >>> weight = module.weight
            >>> bias = module.bias is not None
            >>> stride = module.stride
            >>> padding = module.padding
            >>> dilation = module.dilation
            >>> groups = module.groups
            >>> y = OutputShapeFor(F.conv2d)(x, weight, bias, stride, padding,
            >>>                              dilation, groups)
            >>> print(y)
            >>> y2 = OutputShapeFor(module)(x)
            >>> assert y == y2

            >>> weight = torch.rand(3, 2, 5, 5)
            >>> OutputShapeFor(F.conv2d)((1, 3, 7, 7), weight)
        """
        out_channels, in_channels, kernel_h, kernel_w = weight.shape
        kernel = (kernel_h, kernel_w)
        module = nn.Conv2d(in_channels * groups, out_channels, kernel,
                           bias=bias, stride=stride, padding=padding,
                           dilation=dilation, groups=groups)
        return OutputShapeFor.convnd(module, inputs, 2)

    @staticmethod
    @compute_type(nn.Conv3d)
    def conv3d(module, input_shape):
        return OutputShapeFor.convnd(module, input_shape, 3)

    @staticmethod
    @compute_type(nn.MaxPool1d)
    def maxpool1d(module, input_shape):
        return OutputShapeFor.maxpoolnd(module, input_shape, 1)

    @staticmethod
    @compute_type(nn.MaxPool2d)
    def maxpool2d(module, input_shape):
        return OutputShapeFor.maxpoolnd(module, input_shape, 2)

    @staticmethod
    @compute_type(nn.MaxPool3d)
    def maxpool3d(module, input_shape):
        return OutputShapeFor.maxpoolnd(module, input_shape, 3)

    @staticmethod
    @compute_type(nn.AvgPool1d)
    def avepool1d(module, input_shape):
        return OutputShapeFor.avepoolnd(module, input_shape, 1)

    @staticmethod
    @compute_type(nn.AvgPool2d)
    def avepool2d(module, input_shape):
        return OutputShapeFor.avepoolnd(module, input_shape, 2)

    @staticmethod
    @compute_type(nn.AvgPool3d)
    def avepool3d(module, input_shape):
        return OutputShapeFor.avepoolnd(module, input_shape, 3)

    @staticmethod
    @compute_type(nn.modules.pooling._AdaptiveMaxPoolNd, nn.modules.pooling._AdaptiveAvgPoolNd)
    def adaptive_poolnd(module, input_shape):
        """
        Adaptive pooling is easy because the output-shape is known a-priori
        """
        B, C = input_shape[0:2]
        in_dims = input_shape[2:]

        n = len(in_dims)
        output_dims = ensure_iterablen(module.output_size, n)
        for i, d in enumerate(output_dims):
            if d is None:
                output_dims[i] = in_dims[i]

        output_shape = SHAPE_CLS([B, C] + list(output_dims))
        return output_shape

    @staticmethod
    def convndT(module, input_shape, n):
        r"""
        - Input: :math:`(N, C_{in}, H_{in}, W_{in})`
        - Output: :math:`(N, C_{out}, H_{out}, W_{out})` where
            :math:`H_{out} = (H_{in} - 1) * stride[0] - 2 * padding[0] + kernel\_size[0] + output\_padding[0]`
            :math:`W_{out} = (W_{in} - 1) * stride[1] - 2 * padding[1] + kernel\_size[1] + output\_padding[1]`

        Example:
            >>> from netharn.analytic.output_shape_for import *
            >>> input_shape = (1, 3, 256, 256)
            >>> module = nn.ConvTranspose2d(input_shape[1], 11, kernel_size=2, stride=2)
            >>> output_shape = OutputShapeFor(module)(input_shape)
            >>> print('output_shape = {!r}'.format(output_shape))
            output_shape = (1, 11, 512, 512)

        Example:
            >>> from netharn.analytic.output_shape_for import *
            >>> input_shape = (1, 3, 25, 32, 32)
            >>> module = nn.Conv3d(in_channels=input_shape[1], out_channels=11,
            >>>                    kernel_size=(3, 3, 3), stride=1, padding=0,
            >>>                    dilation=1, groups=1, bias=True)
            >>> output_shape = OutputShapeFor(module)(input_shape)
            >>> print('output_shape = {!r}'.format(output_shape))
            output_shape = (1, 11, 23, 30, 30)
        """
        # N, C_in, *DIMS_in = input_shape
        N, C_in = input_shape[0:2]
        DIMS_in = input_shape[2:]

        if len(DIMS_in) != n:
            raise ValueError('must have {} dims, but got {} '.format(n, len(DIMS_in)))

        C_out = module.out_channels
        stride = module.stride
        kernel_size = module.kernel_size
        output_padding = module.output_padding
        dilation = module.dilation

        padding = module.padding
        DIMS_out = [
            # Fix the docs: https://github.com/pytorch/pytorch/issues/14099
            (D_in - 1) * stride[i] - 2 * padding[i] + (kernel_size[i] - 1) * dilation[i] + output_padding[i] + 1
            for i, D_in in enumerate(DIMS_in)
        ]
        output_shape = SHAPE_CLS([N, C_out] + DIMS_out)
        if math.__name__ == 'sympy':
            output_shape = _simplify(output_shape)
        return output_shape

    @staticmethod
    def convnd(module, input_shape, n):
        r"""
        Notes:
            - Input: :math:`(N, C_{in}, H_{in}, W_{in})`
            - Output: :math:`(N, C_{out}, H_{out}, W_{out})` where
                :math:`H_{out} = floor((H_{in}  + 2 * padding[0] - dilation[0] * (kernel\_size[0] - 1) - 1) / stride[0] + 1)`
                :math:`W_{out} = floor((W_{in}  + 2 * padding[1] - dilation[1] * (kernel\_size[1] - 1) - 1) / stride[1] + 1)`

        Example:
            >>> from netharn.analytic.output_shape_for import *
            >>> input_shape = (1, 3, 256, 256)
            >>> module = nn.Conv2d(input_shape[1], 11, 3, 1, 0)
            >>> output_shape = OutputShapeFor(module)(input_shape)
            >>> print('output_shape = {!r}'.format(output_shape))
            output_shape = (1, 11, 254, 254)
        """
        math = OutputShapeFor.math
        # N, C_in, *DIMS_in = input_shape
        N, C_in = input_shape[0:2]
        DIMS_in = input_shape[2:]

        if len(DIMS_in) != n:
            raise ValueError('must have {} dims, but got {} '.format(n, len(DIMS_in)))

        C_out = module.out_channels
        padding = module.padding
        stride = module.stride
        dilation = module.dilation
        kernel_size = module.kernel_size

        int = builtins.int if math.__name__ == 'math' else ub.identity
        DIMS_out = [
            int(math.floor(
                (D_in + 2 * padding[i] - dilation[i] * (kernel_size[i] - 1) - 1) / stride[i] + 1
            ))
            for i, D_in in enumerate(DIMS_in)
        ]
        output_shape = SHAPE_CLS([N, C_out] + DIMS_out)
        if math.__name__ == 'sympy':
            output_shape = _simplify(output_shape)
        return output_shape

    @staticmethod
    def maxpoolnd(module, input_shape, n):
        r"""
        CommandLine:
            python -m xdoctest netharn.analytic.output_shape_for OutputShapeFor.maxpoolnd:0

        Example:
            >>> from netharn.analytic.output_shape_for import *
            >>> input_shape = (1, 3, 256, 256)
            >>> module = nn.MaxPool2d(kernel_size=2, stride=2)
            >>> output_shape = tuple(OutputShapeFor(module)(input_shape))
            >>> print('output_shape = {!r}'.format(output_shape))
            output_shape = (1, 3, 128, 128)

        Example:
            >>> from netharn.analytic.output_shape_for import *
            >>> input_shape = (1, 512, 37, 37)
            >>> module = nn.MaxPool2d(kernel_size=2, stride=2, padding=0, dilation=1, ceil_mode=True)
            >>> output_shape = tuple(OutputShapeFor(module)(input_shape))
            >>> print('output_shape = {!r}'.format(output_shape))
            output_shape = (1, 512, 19, 19)

        Shape:
            2d Case:
            Same as conv2 forumla except C2 = C1
            - Input: :math:`(N, C, H_{in}, W_{in})`
            - Output: :math:`(N, C, H_{out}, W_{out})` where
            :math:`H_{out} = floor((H_{in}  + 2 * padding[0] - dilation[0] * (kernel\_size[0] - 1) - 1) / stride[0] + 1)`
            :math:`W_{out} = floor((W_{in}  + 2 * padding[1] - dilation[1] * (kernel\_size[1] - 1) - 1) / stride[1] + 1)`
        """
        math = OutputShapeFor.math
        # N, C, *DIMS_in = input_shape
        N, C = input_shape[0:2]
        DIMS_in = input_shape[2:]

        padding = ensure_iterablen(module.padding, n)
        stride = ensure_iterablen(module.stride, n)
        dilation = ensure_iterablen(module.dilation, n)
        kernel_size = ensure_iterablen(module.kernel_size, n)

        trunc = math.ceil if module.ceil_mode else math.floor

        int = builtins.int if math.__name__ == 'math' else ub.identity

        DIMS_out = [
            int(trunc((D_in  + 2 * padding[i] - dilation[i] * (kernel_size[i] - 1) - 1) / stride[i] + 1))
            for i, D_in in enumerate(DIMS_in)
        ]
        output_shape = SHAPE_CLS([N, C] + DIMS_out)
        if math.__name__ == 'sympy':
            output_shape = _simplify(output_shape)
        return output_shape

    @staticmethod
    def avepoolnd(module, input_shape, n):
        r"""
        2D case:
          Shape:
              - Input: :math:`(N, C, H_{in}, W_{in})`
              - Output: :math:`(N, C, H_{out}, W_{out})` where
                :math:`H_{out} = floor((H_{in}  + 2 * padding[0] - kernel\_size[0]) / stride[0] + 1)`
                :math:`W_{out} = floor((W_{in}  + 2 * padding[1] - kernel\_size[1]) / stride[1] + 1)`
        """
        math = OutputShapeFor.math
        # N, C, *DIMS_in = input_shape
        N, C = input_shape[0:2]
        DIMS_in = input_shape[2:]

        padding = ensure_iterablen(module.padding, n)
        stride = ensure_iterablen(module.stride, n)
        kernel_size = ensure_iterablen(module.kernel_size, n)

        int = builtins.int if math.__name__ == 'math' else ub.identity

        DIMS_out = [
            int(math.floor((D_in + 2 * padding[i] - kernel_size[i]) / stride[i] + 1))
            for i, D_in in enumerate(DIMS_in)
        ]
        output_shape = SHAPE_CLS([N, C] + DIMS_out)
        if math.__name__ == 'sympy':
            output_shape = _simplify(output_shape)
        return output_shape

    @staticmethod
    @compute_type(nn.Linear)
    def linear(module, input_shape):
        r"""
           Shape:
               - Input: :math:`(N, *, in\_features)` where `*` means any number of
                 additional dimensions
               - Output: :math:`(N, *, out\_features)` where all but the last dimension
                 are the same shape as the input.
        """
        # N, *other, in_feat = input_shape
        N = input_shape[0]
        other = input_shape[1:-1]
        in_feat = input_shape[-1]  # NOQA

        output_shape = [N] + list(other) + [module.out_features]
        return SHAPE_CLS(output_shape)

    @staticmethod
    def identity(input_shape):
        return SHAPE_CLS(input_shape)

    @staticmethod
    @compute_type(nn.functional.relu)
    def relu_func(input_shape):
        return SHAPE_CLS(input_shape)

    @staticmethod
    @compute_type(nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d,
                  nn.modules.normalization.GroupNorm,
                  nn.modules.normalization.LocalResponseNorm,
                  nn.modules.normalization.LayerNorm, nn.CrossMapLRN2d,
                  nn.InstanceNorm1d, nn.InstanceNorm2d, nn.InstanceNorm3d)
    def normalization(module, input_shape):
        """
            import redbaron
            import torch
            source = open(torch.nn.modules.instancenorm.__file__, 'r').read()
            baron = redbaron.RedBaron(source)
            classes = [item.name for item in baron if item.type == 'class']
            print(', '.join(['nn.{}'.format(c) for c in classes]))

            source = open(torch.nn.modules.normalization.__file__, 'r').read()
            baron = redbaron.RedBaron(source)
            classes = [item.name for item in baron if item.type == 'class']
            print(', '.join(['nn.{}'.format(c) for c in classes]))
        """
        return OutputShapeFor.identity(input_shape)

    @staticmethod
    @compute_type(nn.Dropout, nn.Dropout2d, nn.Dropout3d, nn.AlphaDropout,
                  nn.FeatureAlphaDropout)
    def dropout(module, input_shape):
        return OutputShapeFor.identity(input_shape)

    @staticmethod
    @compute_type(nn.Threshold, nn.RReLU, nn.Hardtanh, nn.ReLU6, nn.ReLU,
                  nn.Sigmoid, nn.Tanh, nn.ELU, nn.CELU, nn.SELU, nn.GLU,
                  nn.Hardshrink, nn.LeakyReLU, nn.LogSigmoid, nn.Softplus,
                  nn.Softshrink, nn.PReLU, nn.Softsign, nn.Tanhshrink,
                  nn.Softmin, nn.Softmax, nn.Softmax2d, nn.LogSoftmax)
    def nonlinearity(module, input_shape):
        r"""
        Ignore:
            import redbaron
            import torch
            source = open(torch.nn.modules.activation.__file__, 'r').read()
            baron = redbaron.RedBaron(source)
            classes = [item.name for item in baron if item.type == 'class']
            print(', '.join(['nn.{}'.format(c) for c in classes]))
        """
        return OutputShapeFor.identity(input_shape)

    @staticmethod
    @compute_type(nn.Sequential)
    def sequential(module, input_shape):
        """
        CommandLine:
            xdoctest -m netharn.analytic.output_shape_for OutputShapeFor.sequential

        Example:
            >>> from netharn.analytic.output_shape_for import *
            >>> self = nn.Sequential(
            >>>     nn.Conv2d(2, 3, kernel_size=3),
            >>>     nn.Conv2d(3, 5, kernel_size=3),
            >>>     nn.Conv2d(5, 7, kernel_size=3),
            >>> )
            >>> shape = OutputShapeFor(self)([1, 1, 7, 11])
            >>> print('shape = {}'.format(ub.repr2(shape, nl=0)))
            >>> print('shape.hidden = {}'.format(ub.repr2(shape.hidden, nl=1)))
            shape = (1, 7, 1, 5)
            shape.hidden = {
                '0': (1, 3, 5, 9),
                '1': (1, 5, 3, 7),
                '2': (1, 7, 1, 5),
            }
        """
        hidden = HiddenShapes()
        shape = input_shape
        for key, child in module._modules.items():
            hidden[key] = shape = OutputShapeFor(child)(shape)
        shape = OutputShape.coerce(shape, hidden=hidden)
        return shape

    @staticmethod
    @compute_type(torchvision.models.resnet.BasicBlock)
    def resent_basic_block(module, input_shape):
        residual_shape = input_shape
        shape = input_shape

        hidden = HiddenShapes()
        hidden['conv1'] = shape = OutputShapeFor(module.conv1)(shape)
        hidden['bn1']   = shape = OutputShapeFor(module.bn1)(shape)
        hidden['relu1'] = shape = OutputShapeFor(module.relu)(shape)

        hidden['conv2'] = shape = OutputShapeFor(module.conv2)(shape)
        hidden['bn2']   = shape = OutputShapeFor(module.bn2)(shape)
        hidden['relu2'] = shape = OutputShapeFor(module.relu)(shape)

        if module.downsample is not None:
            residual_shape = OutputShapeFor(module.downsample)(residual_shape)
            hidden['residual'] = residual_shape

        hidden['join'] = shape
        assert residual_shape[-2:] == shape[-2:], (
            'cannot add residual {} {}'.format(residual_shape, shape))
        shape = OutputShapeFor(module.relu)(shape)
        hidden['relu3'] = shape
        shape = OutputShape.coerce(shape, hidden=hidden)
        return shape

    @staticmethod
    @compute_type(torchvision.models.resnet.Bottleneck)
    def resent_bottleneck(module, input_shape):
        residual_shape = input_shape
        shape = input_shape

        hidden = HiddenShapes()
        hidden['conv1'] = shape = OutputShapeFor(module.conv1)(shape)
        hidden['bn1']   = shape = OutputShapeFor(module.bn1)(shape)
        hidden['relu1'] = shape = OutputShapeFor(module.relu)(shape)

        hidden['conv2'] = shape = OutputShapeFor(module.conv2)(shape)
        hidden['bn2']   = shape = OutputShapeFor(module.bn2)(shape)
        hidden['relu2'] = shape = OutputShapeFor(module.relu)(shape)

        hidden['conv3'] = shape = OutputShapeFor(module.conv3)(shape)
        hidden['bn3']   = shape = OutputShapeFor(module.bn3)(shape)

        if module.downsample is not None:
            residual_shape = OutputShapeFor(module.downsample)(input_shape)
            hidden['residual'] = residual_shape

        assert residual_shape[-2:] == shape[-2:], (
            'cannot add residual {} {}'.format(residual_shape, shape))
        hidden['join'] = shape

        shape = OutputShapeFor(module.relu)(shape)
        hidden['relu3'] = shape

        shape = OutputShape.coerce(shape, hidden=hidden)
        return shape

    @staticmethod
    @compute_type(torchvision.models.resnet.ResNet)
    def resnet_model(module, input_shape):
        """
        Example:
            >>> # xdoctest: +REQUIRES(--network)
            >>> from netharn.analytic.output_shape_for import *
            >>> module = torchvision.models.resnet50()
            >>> input_shape = (1, 3, 224, 224)
            >>> field = OutputShapeFor(module)(input_shape=input_shape)
        """
        shape = input_shape

        hidden = HiddenShapes()
        hidden['conv1'] = shape = OutputShapeFor(module.conv1)(shape)
        hidden['bn1'] = shape = OutputShapeFor(module.bn1)(shape)
        hidden['relu1'] = shape = OutputShapeFor(module.relu)(shape)
        hidden['maxpool'] = shape = OutputShapeFor(module.maxpool)(shape)

        hidden['layer1'] = shape = OutputShapeFor(module.layer1)(shape)
        hidden['layer2'] = shape = OutputShapeFor(module.layer2)(shape)
        hidden['layer3'] = shape = OutputShapeFor(module.layer3)(shape)
        hidden['layer4'] = shape = OutputShapeFor(module.layer4)(shape)

        hidden['avgpool'] = shape = OutputShapeFor(module.avgpool)(shape)

        def prod(args):
            result = args[0]
            for arg in args[1:]:
                result = result * arg
            return result
        shape = (shape[0], prod(shape[1:]))
        hidden['view'] = shape

        hidden['fc'] = shape = OutputShapeFor(module.fc)(shape)
        shape = OutputShape.coerce(shape, hidden=hidden)
        return shape

    @staticmethod
    @compute_type(nn.functional.adaptive_avg_pool2d)
    def adaptive_poolnd_func(input_shape, output_shape):
        """
        Adaptive pooling is easy because the output-shape is known a-priori

        Example:
            >>> from netharn.analytic.output_shape_for import *
            >>> input_shape = (1, 3, 256, 256)
            >>> output_shape = (7, 7)
            >>> output_shape_ = OutputShapeFor(nn.functional.adaptive_avg_pool2d)(input_shape, output_shape)
            >>> print('output_shape = {!r}'.format(output_shape_))
            output_shape = (1, 3, 7, 7)
        """
        B, C = input_shape[0:2]
        in_dims = input_shape[2:]

        n = len(in_dims)
        output_dims = ensure_iterablen(output_shape, n)
        for i, d in enumerate(output_dims):
            if d is None:
                output_dims[i] = in_dims[i]

        output_shape_ = SHAPE_CLS([B, C] + list(output_dims))
        return output_shape_

    @staticmethod
    @compute_type(torch.sigmoid)
    def sigmoid(input_shape):
        return OutputShapeFor.identity(input_shape)

    @staticmethod
    @compute_type(F.pad)
    def pad(x, pad, mode='constant', value=0):
        """
        Example:
            >>> t4d = x = (3, 3, 4, 2)
            >>> pad = p1d = (1, 1)
            >>> out = OutputShapeFor(F.pad)(x, pad)
            >>> print(out)
            (3, 3, 4, 4)
            >>> p2d = (1, 1, 2, 2) # pad last dim by (1, 1) and 2nd to last by (2, 2)
            >>> out = OutputShapeFor.pad(t4d, p2d, "constant", 0)
            >>> print(out)
            (3, 3, 8, 4)
            >>> t4d = (3, 3, 4, 2)
            >>> p3d = (0, 1, 2, 1, 3, 3) # pad by (0, 1), (2, 1), and (3, 3)
            >>> out = OutputShapeFor.pad(t4d, p3d, "constant", 0)
            >>> print(out)
            (3, 9, 7, 3)
        """
        new_x = list(x)
        dim = len(new_x)
        for idx, dpad in enumerate(ub.chunks(pad, 2), start=1):
            dimx = dim - idx
            lpad, rpad = dpad
            new_x[dimx] = x[dimx] + lpad + rpad
        out = SHAPE_CLS(new_x)
        return out

    @staticmethod
    @compute_type(torch.cat)
    def cat(input_shapes, dim=0):
        """
        Example:
            >>> from netharn.analytic.output_shape_for import *
            >>> input_shape1 = (1, 3, 256, 256)
            >>> input_shape2 = (1, 4, 256, 256)
            >>> input_shapes = [input_shape1, input_shape2]
            >>> output_shape = OutputShapeFor(torch.cat)(input_shapes, dim=1)
            >>> print('output_shape = {!r}'.format(output_shape))
            output_shape = (1, 7, 256, 256)
        """
        n_dims = max(map(len, input_shapes))
        assert n_dims == min(map(len, input_shapes))
        output_shape = [None] * n_dims
        for shape in input_shapes:
            for i, v in enumerate(shape):
                if output_shape[i] is None:
                    output_shape[i] = v
                else:
                    if i == dim:
                        output_shape[i] += v
                    else:
                        assert output_shape[i] == v, 'inconsistent dims {}'.format(input_shapes)
        return SHAPE_CLS(output_shape)

    @staticmethod
    @compute_type(DataSerial)
    def data_serial(module, *args, **kw):
        return OutputShapeFor(module.module)(*args, **kw)

    @staticmethod
    @compute_type(torch.nn.DataParallel)
    def data_parallel(module, *args, **kw):
        return OutputShapeFor(module.module)(*args, **kw)

    @staticmethod
    def getitem(arr):
        """
        Wraps getitem calls

        Example:
            >>> arr = (2, 32, 9, 9)
            >>> result = OutputShapeFor.getitem(arr)[:, 0:4]
            >>> assert result == [2, 4, 9, 9]
        """
        return _ShapeGetItem(arr)

    @staticmethod
    def view(arr, *args):
        """
        Wraps view calls

        Example:
            >>> arr = (2, 32, 9, 9)
            >>> result = OutputShapeFor.view(arr, -1)
            >>> assert result == (5184,)
        """
        from netharn import layers
        reshape = layers.Reshape(*args)
        return reshape.output_shape_for(arr)

    @staticmethod
    def shape(arr):
        """
        Wraps shape calls

        Example:
            >>> arr = (2, 32, 9, 9)
            >>> result = OutputShapeFor.shape(arr)
            >>> assert result == arr
        """
        return arr

    @staticmethod
    def add(arr1, arr2):
        return _output_shape_broadcast(arr1, arr2)

    @staticmethod
    def mul(arr1, arr2):
        return _output_shape_broadcast(arr1, arr2)

    @staticmethod
    def sub(arr1, arr2):
        return _output_shape_broadcast(arr1, arr2)

    @staticmethod
    def div(arr1, arr2):
        return _output_shape_broadcast(arr1, arr2)


def _output_shape_broadcast(arr1, arr2):
    """
    Args:
        arr1 (Tuple | scalar): shape of arr1 or a scalar
        arr2 (Tuple | scalar): shape of arr2 or a scalar
    """
    if not ub.iterable(arr1):
        return arr2
    if not ub.iterable(arr2):
        return arr1
    if tuple(arr1) != tuple(arr2):

        if len(arr1) == len(arr2):
            arr3 = []
            for d1, d2 in zip(arr1, arr2):
                if d1 is None or d1 < 0:
                    raise NotImplementedError
                if d2 is None or d2 < 0:
                    raise NotImplementedError
                if d1 == d2:
                    arr3.append(d1)
                elif d1 == 1:
                    arr3.append(d2)
                elif d2 == 1:
                    arr3.append(d1)
                else:
                    raise ValueError('broadcast seems bad')
            arr3 = type(arr1)(arr3)
            return arr3

        # TODO: handle broadcast
        raise NotImplementedError('Full broadcast not implemented {} != {}'.format(arr1, arr2))
    return arr1


class _ShapeGetItem(object):
    def __init__(self, inp):
        self.inp = inp

    def __getitem__(self, slices):
        ellipsis_type = type(Ellipsis)
        oup = list(self.inp)
        if isinstance(slices, slice):
            slices = (slices,)

        if isinstance(slices, tuple):
            for i, sl in enumerate(slices):
                if isinstance(sl, ellipsis_type):
                    assert i == len(slices) - 1
                    break
                start, stop, step = sl.indices(oup[i])
                oup[i] = (stop - start) // step
        return oup


def ensure_iterablen(scalar, n):
    try:
        iter(scalar)
    except TypeError:
        return [scalar] * n
    return scalar



from netharn.layers import common
from netharn.layers import rectify
from netharn.layers import conv_norm
import numpy as np


class MultiLayerPerceptronNd(common.Module):
    """
    A multi-layer perceptron network for n dimensional data

    Choose the number and size of the hidden layers, number of output channels,
    wheather to user residual connections or not, nonlinearity, normalization,
    dropout, and more.

    Args:
        dim (int): specify if the data is 0, 1, 2, 3, or 4 dimensional.
        in_channels (int):
        hidden_channels (List[int]): or an int specifying the number of hidden
            layers (we choose the channel size to linearly interpolate between
            input and output channels)
        out_channels (int):
        dropout (float, default=0): amount of dropout to use
        norm (str, default='batch'): type of normalization layer (e.g. batch or group)
        noli (str, default='relu'): type of nonlinearity
        residual (bool, default=False):
            if true includes a resitual skip connection between inputs and
            outputs.

    CommandLine:
        xdoctest -m ~/code/netharn/netharn/layers/perceptron.py MultiLayerPerceptronNd:0

    Example:
        >>> from netharn.layers.perceptron import *
        >>> import ubelt as ub
        >>> self = MultiLayerPerceptronNd(dim=1, in_channels=128, hidden_channels=3, out_channels=2)
        >>> print(self)
        MultiLayerPerceptronNd...
        >>> shape = self.output_shape_for([1, 128, 7])
        >>> print('shape = {!r}'.format(shape))
        >>> print('shape.hidden = {}'.format(ub.repr2(shape.hidden, nl=1)))
        shape = (1, 2, 7)
        shape.hidden = {
            'hidden0': {'conv': (1, 96, 7), 'norm': (1, 96, 7), 'noli': (1, 96, 7)},
            'dropout0': (1, 96, 7),
            'hidden1': {'conv': (1, 65, 7), 'norm': (1, 65, 7), 'noli': (1, 65, 7)},
            'dropout1': (1, 65, 7),
            'hidden2': {'conv': (1, 34, 7), 'norm': (1, 34, 7), 'noli': (1, 34, 7)},
            'dropout2': (1, 34, 7),
            'output': (1, 2, 7),
        }
        >>> import netharn as nh
        >>> nh.OutputShapeFor(self)._check_consistency([1, 128, 7])
        (1, 2, 7)
        >>> print('self._hidden_channels = {!r}'.format(self._hidden_channels))

    Example:
        >>> from netharn.layers.perceptron import *
        >>> import ubelt as ub
        >>> self = MultiLayerPerceptronNd(0, 128, [256, 64], residual=True,
        >>>                               norm='group', out_channels=2)
        >>> print(self)
        >>> input_shape = (None, 128)
        >>> print(ub.repr2(self.output_shape_for(input_shape).hidden, nl=-1))

    Example:
        >>> from netharn.layers.perceptron import *
        >>> import ubelt as ub
        >>> self = MultiLayerPerceptronNd(0, 128, [], residual=False,
        >>>                               norm='group', out_channels=2)
        >>> print(self)
        >>> input_shape = (None, 128)
        >>> print(ub.repr2(self.output_shape_for(input_shape).hidden, nl=-1))
    """
    def __init__(self, dim, in_channels, hidden_channels, out_channels,
                 bias=True, dropout=0, noli='relu', norm='batch',
                 residual=False):

        super(MultiLayerPerceptronNd, self).__init__()
        dropout_cls = rectify.rectify_dropout(dim)
        conv_cls = rectify.rectify_conv(dim=dim)
        curr_in = in_channels

        if isinstance(hidden_channels, int):
            n = hidden_channels
            hidden_channels = np.linspace(in_channels, out_channels, n + 1,
                                          endpoint=False)[1:]
            hidden_channels = hidden_channels.round().astype(np.int).tolist()
        self._hidden_channels = hidden_channels

        hidden = self.hidden = common.Sequential()
        for i, curr_out in enumerate(hidden_channels):
            layer = conv_norm.ConvNormNd(dim, curr_in, curr_out, kernel_size=1,
                                         bias=False, noli=noli, norm=norm)
            hidden.add_module('hidden{}'.format(i), layer)
            hidden.add_module('dropout{}'.format(i), dropout_cls(p=dropout))
            curr_in = curr_out

        outkw = {'bias': bias}
        if dim > 0:
            outkw['kernel_size'] = 1
        self.hidden.add_module('output', conv_cls(curr_in, out_channels, **outkw))

        if residual:
            self.skip = conv_cls(in_channels, out_channels, **outkw)
        else:
            self.skip = None

        self.in_channels = in_channels
        self.out_channels = out_channels

    def forward(self, inputs):
        outputs = self.hidden(inputs)
        if self.skip:
            outputs = self.skip(inputs) + outputs
        return outputs

    def output_shape_for(self, input_shape):
        outputs = self.hidden.output_shape_for(input_shape)
        if self.skip:
            import netharn as nh
            skip = nh.OutputShapeFor(self.skip)(input_shape)
            outputs.hidden['skip'] = skip
        return outputs

    def receptive_field_for(self, input_field=None):
        import netharn as nh
        field = nh.ReceptiveFieldFor(self.hidden)(input_field)
        if self.skip:
            skip = nh.ReceptiveFieldFor(self.skip)(field)
            field.hidden['skip'] = skip
        return field



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
import numpy as np
import cv2
import six
from imgaug.parameters import (Uniform, Binomial)
from netharn.data.transforms import augmenter_base
import kwimage
import imgaug


def demodata_hsv_image(w=200, h=200):
    """
    Example:
        >>> rgb255 = demodata_hsv_image()
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.figure(doclf=True, fnum=1)
        >>> kwplot.imshow(rgb255, colorspace='rgb')
        >>> kwplot.show_if_requested()
    """

    hsv = np.zeros((h, w, 3), dtype=np.float32)

    hue = np.linspace(0, 360, num=w)
    hsv[:, :, 0] = hue[None, :]

    sat = np.linspace(0, 1, num=h)
    hsv[:, :, 1] = sat[:, None]

    val = np.linspace(0, 1, num=3)
    parts = []
    for v in val:
        p = hsv.copy()
        p[:, :, 2] = v
        parts.append(p)
    final_hsv = np.hstack(parts)
    rgb01 = cv2.cvtColor(final_hsv, cv2.COLOR_HSV2RGB)
    rgb255 = (rgb01 * 255).astype(np.uint8)
    return rgb255


class HSVShift(augmenter_base.ParamatarizedAugmenter):
    """
    Perform random HSV shift on the RGB data.

    MODIFIED FROM LIGHTNET YOLO into imgaug format

    Args:
        hue (Number): Random number between -hue,hue is used to shift the hue.
            The number is specified as a percentage of the available hue space
            (e.g. hue * 255 or hue * 360).
        saturation (Number): Random number between 1,saturation is used to
            scale the saturation; 50% chance to get 1/dSaturation instead of
            dSaturation
        value (Number): Random number between 1,value is used to scale the
            value; 50% chance to get 1/dValue in stead of dValue

        shift_sat (Number, default=0): random shift applied to saturation
        shift_val (Number, default=0): random shift applied to value

    CommandLine:
        python -m netharn.data.transforms.augmenters HSVShift --show

    Example:
        >>> self = HSVShift(0.1, 1.5, 1.5)
        >>> img = demodata_hsv_image()
        >>> aug = self.augment_image(img)
        >>> det = self.to_deterministic()
        >>> assert np.all(det.augment_image(img) == det.augment_image(img))
        >>> # xdoc: +REQUIRES(--show)
        >>> import kwplot
        >>> import ubelt as ub
        >>> kwplot.autompl()
        >>> kwplot.figure(doclf=True, fnum=3)
        >>> self = HSVShift(0.5, 1.5, 1.5, shift_sat=1.0, shift_val=1.0)
        >>> pnums = kwplot.PlotNums(5, 5)
        >>> #random_state = self.random_state
        >>> import kwarray
        >>> self.reseed(kwarray.ensure_rng(0))
        >>> kwplot.imshow(img, colorspace='rgb', pnum=pnums[0], title='orig')
        >>> for i in range(1, len(pnums)):
        >>>     aug = self.augment_image(img)
        >>>     title = 'aug: {}'.format(ub.repr2(self._prev_params, nl=0, precision=3))
        >>>     kwplot.imshow(aug, colorspace='rgb', pnum=pnums[i], title=title)
        >>> kwplot.show_if_requested()

    Ignore:
        >>> from netharn.data.transforms.augmenters import *
        >>> lnpre = ub.import_module_from_path(ub.expandpath('~/code/lightnet/lightnet/data/transform/_preprocess.py'))
        >>> self = lnpre.HSVShift(0.1, 1.5, 1.5)
        >>> from PIL import Image
        >>> img = demodata_hsv_image()
        >>> from_ = ub.identity
        >>> #img = Image.fromarray(img)
        >>> #from_ = np.array
        >>> aug = self(img)
        >>> # xdoc: +REQUIRES(--show)
        >>> import kwplot
        >>> import ubelt as ub
        >>> kwplot.autompl()
        >>> kwplot.figure(doclf=True, fnum=1)
        >>> import random
        >>> random.seed(0)
        >>> pnums = kwplot.PlotNums(5, 5)
        >>> kwplot.imshow(from_(img), colorspace='rgb', pnum=pnums[0], title='orig')
        >>> for i in range(1, len(pnums)):
        >>>     aug = self(img)
        >>>     #title = 'aug: {}'.format(ub.repr2(self._prev_params, nl=0, precision=3))
        >>>     title = 'foo'
        >>>     kwplot.imshow(from_(aug), colorspace='rgb', pnum=pnums[i], title=title)
        >>> kwplot.show_if_requested()
    """
    def __init__(self, hue, sat, val, shift_sat=0, shift_val=0,
                 input_colorspace='rgb'):
        super(HSVShift, self).__init__()
        self.input_colorspace = input_colorspace
        self.hue = Uniform(-hue, hue)
        self.sat = Uniform(1, sat)
        self.val = Uniform(1, val)

        self.shift_sat = Uniform(-shift_sat, shift_sat)
        self.shift_val = Uniform(-shift_val, shift_val)

        self.flip_val = Binomial(.5)
        self.flip_sat = Binomial(.5)

    def _augment_heatmaps(self, heatmaps_on_images, *args, **kw):
        return heatmaps_on_images

    def _augment_images(self, images, random_state, parents, hooks):
        return [self.forward(img, random_state) for img in images]

    def _augment_keypoints(self, keypoints_on_images, random_state, parents,
                           hooks):
        return keypoints_on_images

    def forward(self, img, random_state=None):
        assert self.input_colorspace == 'rgb'
        assert img.dtype.kind == 'u' and img.dtype.itemsize == 1

        dh = self.hue.draw_sample(random_state)
        ds = self.sat.draw_sample(random_state)
        dv = self.val.draw_sample(random_state)

        shift_s = self.shift_sat.draw_sample(random_state)
        shift_v = self.shift_val.draw_sample(random_state)

        if self.flip_sat.draw_sample(random_state):
            ds = 1.0 / ds
        if self.flip_val.draw_sample(random_state):
            dv = 1.0 / dv

        self._prev_params = (dh, ds, dv)

        # Note the cv2 conversion to HSV does not go into the 0-1 range,
        # instead it goes into (0-360, 0-1, 0-1) for hue, sat, and val.
        img01 = img.astype(np.float32) / 255.0
        hsv = cv2.cvtColor(img01, cv2.COLOR_RGB2HSV)

        hue_bound = 360.0
        sat_bound = 1.0
        val_bound = 1.0

        def wrap_hue(new_hue, hue_bound):
            """ This is about 10x faster than using modulus """
            out = new_hue
            out[out >= hue_bound] -= hue_bound
            out[out < 0] += hue_bound
            return out

        # add to hue
        hsv[:, :, 0] = wrap_hue(hsv[:, :, 0] + (hue_bound * dh), hue_bound)
        if shift_s != 0:
            # shift saturation and value
            hsv[:, :, 1] = np.clip(shift_s + hsv[:, :, 1], 0.0, sat_bound)
        if shift_v != 0:
            hsv[:, :, 2] = np.clip(shift_v + hsv[:, :, 2], 0.0, val_bound)
        # scale saturation and value
        hsv[:, :, 1] = np.clip(ds * hsv[:, :, 1], 0.0, sat_bound)
        hsv[:, :, 2] = np.clip(dv * hsv[:, :, 2], 0.0, val_bound)

        img01 = cv2.cvtColor(hsv, cv2.COLOR_HSV2RGB)
        img255 = (img01 * 255).astype(np.uint8)
        return img255


class Resize(augmenter_base.ParamatarizedAugmenter):
    """
    Transform images and annotations to the right network dimensions.

    Args:
        target_size (tuple): Scale images to this size (w, h) keeping the
        aspect ratio using a letterbox.

    CommandLine:
        xdoctest -m /home/joncrall/code/netharn/netharn/data/transforms/augmenters.py Resize:0

    Example:
        >>> from netharn.data.transforms.augmenters import *  # NOQA
        >>> img = demodata_hsv_image()
        >>> box = kwimage.Boxes([[.45, .05, .10, .05], [0., 0.0, .199, .199], [.24, .05, .01, .05]], format='xywh').to_tlbr()
        >>> bboi = box.to_imgaug(shape=tuple(img.shape))
        >>> self = Resize((40, 30))
        >>> aug1  = self.augment_image(img)
        >>> bboi1 = self.augment_bounding_boxes([bboi])[0]
        >>> box1 = kwimage.Boxes.from_imgaug(bboi1)
        >>> assert box1.br_y.max()  > 8
        >>> assert tuple(bboi1.shape[0:2]) == (30, 40)
        >>> assert tuple(aug1.shape[0:2]) == (30, 40)

    Example:
        >>> from netharn.data.transforms.augmenters import *  # NOQA
        >>> img = demodata_hsv_image()
        >>> box = kwimage.Boxes([[450, 50, 100, 50], [0.0, 0, 199, 199], [240, 50, 10, 50]], format='xywh').to_tlbr()
        >>> bboi = box.to_imgaug(shape=img.shape)
        >>> imgT = np.ascontiguousarray(img.transpose(1, 0, 2))
        >>> bboiT = box.transpose().to_imgaug(shape=imgT.shape)
        >>> self = Resize((40, 30))
        >>> self2 = Resize((1000, 1000))
        >>> # ---------------------------
        >>> aug1  = self.augment_image(img)
        >>> bboi1 = self.augment_bounding_boxes([bboi])[0]
        >>> aug2  = self.augment_image(imgT)
        >>> bboi2 = self.augment_bounding_boxes([bboiT])[0]
        >>> aug3  = self2.augment_image(img)
        >>> bboi3 = self2.augment_bounding_boxes([bboi])[0]
        >>> aug4  = self2.augment_image(imgT)
        >>> bboi4 = self2.augment_bounding_boxes([bboiT])[0]
        >>> # ---------------------------
        >>> # xdoc: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.figure(doclf=True, fnum=1)
        >>> pnum_ = kwplot.PlotNums(3, 2)
        >>> kwplot.imshow(img, colorspace='rgb', pnum=pnum_(), title='orig')
        >>> kwplot.draw_boxes(kwimage.Boxes.from_imgaug(bboi))
        >>> kwplot.imshow(imgT, colorspace='rgb', pnum=pnum_(), title='origT')
        >>> kwplot.draw_boxes(kwimage.Boxes.from_imgaug(bboiT))
        >>> # ----
        >>> kwplot.imshow(aug1, colorspace='rgb', pnum=pnum_())
        >>> #kwplot.draw_boxes(kwimage.Boxes.from_imgaug(bboi1))
        >>> x = kwimage.Boxes.from_imgaug(bboi1).translate((-0.5, -0.5))
        >>> kwplot.draw_boxes(x)
        >>> kwplot.imshow(aug2, colorspace='rgb', pnum=pnum_())
        >>> x = kwimage.Boxes.from_imgaug(bboi2).translate((-0.5, -0.5))
        >>> kwplot.draw_boxes(x)
        >>> #kwplot.draw_boxes(kwimage.Boxes.from_imgaug(bboi2))
        >>> # ----
        >>> kwplot.imshow(aug3, colorspace='rgb', pnum=pnum_())
        >>> kwplot.draw_boxes(kwimage.Boxes.from_imgaug(bboi3))
        >>> kwplot.imshow(aug4, colorspace='rgb', pnum=pnum_())
        >>> kwplot.draw_boxes(kwimage.Boxes.from_imgaug(bboi4))

    Ignore:
        image = img
        target_size = np.array(self.target_size)
        orig_size = np.array(img.shape[0:2][::-1])
        shift, scale, embed_size = self._letterbox_transform(orig_size,
                                                             target_size)
    """
    def __init__(self, target_size, fill_color=127, mode='letterbox',
                 border='constant', random_state=None):
        super(Resize, self).__init__(random_state=random_state)
        self.target_size = None if target_size is None else np.array(target_size)
        self.mode = mode

        import imgaug.parameters as iap
        if fill_color == imgaug.ALL:
            self.fill_color = iap.Uniform(0, 255)
        else:
            self.fill_color = iap.handle_continuous_param(
                fill_color, "fill_color", value_range=None,
                tuple_to_uniform=True, list_to_choice=True)

        self._cv2_border_type_map = {
            'constant': cv2.BORDER_CONSTANT,
            'edge': cv2.BORDER_REPLICATE,
            'linear_ramp': None,
            'maximum': None,
            'mean': None,
            'median': None,
            'minimum': None,
            'reflect': cv2.BORDER_REFLECT_101,
            'symmetric': cv2.BORDER_REFLECT,
            'wrap': cv2.BORDER_WRAP,
            cv2.BORDER_CONSTANT: cv2.BORDER_CONSTANT,
            cv2.BORDER_REPLICATE: cv2.BORDER_REPLICATE,
            cv2.BORDER_REFLECT_101: cv2.BORDER_REFLECT_101,
            cv2.BORDER_REFLECT: cv2.BORDER_REFLECT
        }
        if isinstance(border, six.string_types):
            if border == imgaug.ALL:
                border = [k for k, v in self._cv2_border_type_map.items()
                          if v is not None and isinstance(k, six.string_types)]
            else:
                border = [border]
        if isinstance(border, (list, tuple)):
            from imgaug.parameters import Choice
            border = Choice(border)
        self.border = border
        assert self.mode == 'letterbox', 'thats all folks'

    def forward(self, img, random_state=None):
        orig_size = np.array(img.shape[0:2][::-1])
        assert self.mode == 'letterbox', 'thats all folks'
        shift, scale, embed_size = self._letterbox_transform(orig_size,
                                                             self.target_size)
        new_img = self._img_letterbox_apply(img, embed_size, shift,
                                            self.target_size)
        return new_img

    def _augment_bounding_boxes(self, bounding_boxes_on_images, random_state,
                                parents, hooks):
        # Fix for imgaug 0.4.0
        return self._augment_bounding_boxes_as_keypoints(
            bounding_boxes_on_images, random_state, parents, hooks)

    def _augment_polygons(self, polygons_on_images, random_state, parents,
                          hooks):
        # Fix for imgaug 0.4.0
        return self._augment_polygons_boxes_as_keypoints(
            polygons_on_images, random_state, parents, hooks)

    def _augment_images(self, images, random_state, parents, hooks):
        self.target_size = None if self.target_size is None else np.array(self.target_size)
        return [self.forward(img, random_state) for img in images]

    def _augment_keypoints(self, keypoints_on_images, random_state, parents,
                           hooks):
        """
        Example:
            >>> import imgaug
            >>> tlbr = [[0, 0, 10, 10], [1, 2, 8, 9]]
            >>> shape = (20, 40, 3)
            >>> bboi = kwimage.Boxes(tlbr, 'tlbr').to_imgaug(shape)
            >>> bounding_boxes_on_images = [bboi]
            >>> kps_ois = []
            >>> for bbs_oi in bounding_boxes_on_images:
            >>>     kps = []
            >>>     for bb in bbs_oi.bounding_boxes:
            >>>         kps.extend(bb.to_keypoints())
            >>>     kps_ois.append(imgaug.KeypointsOnImage(kps, shape=bbs_oi.shape))
            >>> keypoints_on_images = kps_ois
            >>> self = LetterboxResize((400, 400))
            >>> aug = self.augment_keypoints(keypoints_on_images)
            >>> assert np.all(aug[0].shape == self.target_size[::-1])
        """
        result = []
        target_size = np.array(self.target_size)
        target_shape = target_size[::-1]
        prev_size = None
        for i, keypoints_on_image in enumerate(keypoints_on_images):
            orig_size = (keypoints_on_image.width, keypoints_on_image.height)

            if prev_size != orig_size:
                # Cache previously computed values
                shift, scale, embed_size = self._letterbox_transform(
                    orig_size, target_size)
                prev_size = orig_size

            try:
                xy = keypoints_on_image.to_xy_array()
            except (Exception, AttributeError):
                xy = keypoints_on_image.get_coords_array()
            xy_aug = (xy * scale) + shift

            try:
                new_keypoint = imgaug.KeypointsOnImage.from_xy_array(
                    xy_aug, shape=target_shape)
            except (Exception, AttributeError):
                new_keypoint = imgaug.KeypointsOnImage.from_coords_array(
                    xy_aug, shape=target_shape)
            # Fix bug in imgaug (TODO: report the bug)
            new_keypoint.shape = target_shape
            result.append(new_keypoint)
        return result

    def _img_letterbox_invert(self, img, orig_size, target_size):
        """
        Args:
            img : the image to scale back up
            orig_size : original wh of the image
            target_size : network input wh

        Example:
            >>> orig_img = demodata_hsv_image(w=100, h=200)
            >>> orig_size = np.array(orig_img.shape[0:2][::-1])
            >>> target_size = (416, 416)
            >>> self = Resize(target_size)
            >>> img = self.forward(orig_img)
            >>> inverted_img = self._img_letterbox_invert(img, orig_size, target_size)
            >>> assert inverted_img.shape == orig_img.shape
            >>> # xdoc: +REQUIRES(--show)
            >>> import kwplot
            >>> kwplot.imshow(orig_img, fnum=1, pnum=(1, 3, 1))
            >>> kwplot.imshow(img, fnum=1, pnum=(1, 3, 2))
            >>> kwplot.imshow(inverted_img, fnum=1, pnum=(1, 3, 3))
        """
        shift, scale, embed_size = self._letterbox_transform(orig_size, target_size)
        top, bot, left, right = self._padding(embed_size, shift, target_size)

        # Undo padding
        h, w = img.shape[0:2]
        unpadded_img = img[top:h - bot, left:w - right]

        sf = orig_size / embed_size
        dsize = tuple(orig_size)
        # Choose INTER_AREA if we are shrinking the image
        interpolation = cv2.INTER_AREA if sf.sum() < 2 else cv2.INTER_CUBIC
        inverted_img = cv2.resize(unpadded_img, dsize, interpolation=interpolation)
        return inverted_img

    def _boxes_letterbox_apply(self, boxes, orig_size, target_size):
        """
        Apply the letterbox transform to these bounding boxes

        """
        shift, scale, embed_size = self._letterbox_transform(orig_size, target_size)
        new_boxes = boxes.scale(scale).translate(shift)
        return new_boxes

    def _boxes_letterbox_invert(self, boxes, orig_size, target_size):
        """
        Undo the letterbox transform for these bounding boxes. Moves
        the box from `target_size` coordinatse (which are probably square)
        to `orig_size` coordinates (which are probably not square).

        Args:
            boxes (kwimage.Boxes) : boxes to rework in `target_size` coordinates
            orig_size : original wh of the image
            target_size : network input wh (i.e. inp_size)

        Example:
            >>> target_size = (416, 416)
            >>> orig_size = (1000, 400)
            >>> cxywh_norm = kwimage.Boxes(np.array([[.5, .5, .2, .2]]), 'cxywh')
            >>> self = Resize(target_size)
            >>> cxywh = self._boxes_letterbox_invert(cxywh_norm, orig_size, target_size)
            >>> cxywh_norm2 = self._boxes_letterbox_apply(cxywh, orig_size, target_size)
            >>> assert np.allclose(cxywh_norm2.data, cxywh_norm.data)
        """
        shift, scale, embed_size = self._letterbox_transform(orig_size, target_size)
        new_boxes = boxes.translate(-shift).scale(1.0 / scale)
        return new_boxes

    def _letterbox_transform(self, orig_size, target_size):
        """
        aspect ratio preserving scaling + extra padding to equal target size

        scale should be applied before the shift.

        Args:
            orig_size : original wh of the image
            target_size : network input wh

        Returns:
            Tuple:
                shift: x,y shift
                scale: w,h scale
                embed_size: innner w,h of unpadded region

        Example:
            >>> # xdoctest: +IGNORE_WHITESPACE
            >>> Resize(None)._letterbox_transform([5, 10], [10, 10])
            (array([2, 0]), array([1., 1.]), array([ 5, 10]))
            >>> Resize(None)._letterbox_transform([10, 5], [10, 10])
            (array([0, 2]), array([1., 1.]), array([10,  5]))
        """
        # determine if width or the height should be used as the scale factor.
        orig_size = np.array(orig_size)
        target_size = np.array(target_size)
        fw, fh = orig_size / target_size
        sf = 1 / fw if fw >= fh else 1 / fh

        # Whats the closest integer size we can resize to?
        embed_size = np.round(orig_size * sf).astype(np.int)
        # Determine how much padding we need for the top/left side
        # Note: the right/bottom side might need an extra pixel of padding
        # depending on rounding issues.
        shift = np.round((target_size - embed_size) / 2).astype(np.int)

        scale = embed_size / orig_size
        return shift, scale, embed_size

    @staticmethod
    def _padding(embed_size, shift, target_size):
        pad_lefttop = shift
        pad_rightbot = target_size - (embed_size + shift)

        left, top = pad_lefttop
        right, bot = pad_rightbot
        return top, bot, left, right

    def _img_letterbox_apply(self, img, embed_size, shift, target_size):
        top, bot, left, right = self._padding(embed_size, shift, target_size)

        orig_size = np.array(img.shape[0:2][::-1])
        channels = kwimage.num_channels(img)

        sf = embed_size / orig_size
        dsize = tuple(embed_size)
        # Choose INTER_AREA if we are shrinking the image
        interpolation = cv2.INTER_AREA if sf.sum() < 2 else cv2.INTER_CUBIC
        if any(d < 0 for d in dsize):
            raise ValueError('dsize={} must be non-negative'.format(dsize))
        scaled = cv2.resize(img, dsize, interpolation=interpolation)

        border = self.border.draw_sample()
        cval = self.fill_color.draw_sample()

        if scaled.dtype == 'f':
            value = (float(cval),) * channels
        else:
            value = (int(cval),) * channels

        borderType = self._cv2_border_type_map[border]
        if borderType is None:
            raise ValueError('bad border type border={}, borderType={}'.format(
                border, borderType))

        hwc255 = cv2.copyMakeBorder(scaled, top, bot, left, right,
                                    borderType=borderType,
                                    value=value)
        return hwc255


LetterboxResize = Resize

if __name__ == '__main__':
    """
    CommandLine:
        python -m netharn.data.transforms.augmenters all
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



# -*- coding: utf-8 -*-
"""
Deployment component of the Pytorch exporter.

This file contains DeployedModel, which consists of logic to take the
model topology definition along with the "best" snapshot in a training
directory and package it up into a standalone zipfile. The DeployedModel can
also be used to reload model from this zipfile. Thus this zipfile can be passed
around as a pytorch model topology+pretrained weights transfer format.

The following docstring illustrates how this module may be used.

CommandLine:
    # Runs the following example
    xdoctest -m netharn.export.deployer __doc__:0

    # Runs all the doctests
    xdoctest -m netharn.export.deployer all

Example:
    >>> # xdoc: +IGNORE_WANT
    >>> # This example will train a small model and then deploy it.
    >>> import netharn as nh
    >>> #
    >>> #################################################
    >>> print('--- STEP 1: TRAIN A MODEL ---')
    >>> # This will train a toy model with toy data using netharn
    >>> hyper = nh.HyperParams(**{
    >>>     'workdir'     : ub.ensure_app_cache_dir('netharn/tests/deploy'),
    >>>     'nice'        : 'deploy_demo',
    >>>     'xpu'         : nh.XPU.coerce('cpu'),
    >>>     'datasets'    : {
    >>>         'train': nh.data.ToyData2d(size=3, border=1, n=256, rng=0),
    >>>         'test':  nh.data.ToyData2d(size=3, border=1, n=128, rng=1),
    >>>     },
    >>>     'loaders'     : {'batch_size': 64},
    >>>     'model'       : (nh.models.ToyNet2d, {}),
    >>>     'optimizer'   : (nh.optimizers.SGD, {'lr': 0.0001}),
    >>>     'criterion'   : (nh.criterions.CrossEntropyLoss, {}),
    >>>     'initializer' : (nh.initializers.KaimingNormal, {}),
    >>>     'scheduler'   : (nh.schedulers.ListedLR, {
    >>>         'points': {0: .01, 3: 0.1},
    >>>         'interpolate': True,
    >>>     }),
    >>>     'monitor'     : (nh.Monitor, {'max_epoch': 3,}),
    >>> })
    >>> harn = nh.FitHarn(hyper)
    >>> harn.preferences['use_tensorboard'] = False
    >>> harn.preferences['timeout'] = 1
    >>> harn.intervals['test'] = 1
    >>> harn.initialize(reset='delete')
    >>> harn.run()
    --- STEP 1: TRAIN A MODEL ---
    RESET HARNESS BY DELETING EVERYTHING IN TRAINING DIR
    Symlink: .../.cache/netharn/tests/deploy/fit/runs/deploy_demo/onnxqaww -> .../.cache/netharn/tests/deploy/_mru
    ......
    Symlink: .../.cache/netharn/tests/deploy/fit/runs/deploy_demo/onnxqaww -> .../.cache/netharn/tests/deploy/fit/nice/deploy_demo
    ......
    INFO: Model has 824 parameters
    INFO: Mounting ToyNet2d model on CPU
    INFO: Exported model topology to .../.cache/netharn/tests/deploy/fit/runs/deploy_demo/onnxqaww/ToyNet2d_2a3f49.py
    INFO: Initializing model weights with: <netharn.initializers.nninit_core.KaimingNormal object at 0x7fc67efdf8d0>
    INFO:  * harn.train_dpath = '.../.cache/netharn/tests/deploy/fit/runs/deploy_demo/onnxqaww'
    INFO:  * harn.nice_dpath  = '.../.cache/netharn/tests/deploy/fit/nice/deploy_demo'
    INFO: Snapshots will save to harn.snapshot_dpath = '.../.cache/netharn/tests/deploy/fit/runs/deploy_demo/onnxqaww/torch_snapshots'
    INFO: ARGV:
        .../.local/conda/envs/py36/bin/python .../.local/conda/envs/py36/bin/ipython
    INFO: === begin training 0 / 3 : deploy_demo ===
    epoch lr:0.01 │ vloss is unevaluated 0/3... rate=0 Hz, eta=?, total=0:00:00, wall=19:32 EST
    train loss:0.717 │ 100.00% of 64x8... rate=2093.02 Hz, eta=0:00:00, total=0:00:00, wall=19:32 EST
    test loss:0.674 │ 100.00% of 64x4... rate=14103.48 Hz, eta=0:00:00, total=0:00:00, wall=19:32 EST
    Populating the interactive namespace from numpy and matplotlib
    INFO: === finish epoch 0 / 3 : deploy_demo ===
    epoch lr:0.04 │ vloss is unevaluated 1/3... rate=0.87 Hz, eta=0:00:02, total=0:00:01, wall=19:32 EST
    train loss:0.712 │ 100.00% of 64x8... rate=2771.29 Hz, eta=0:00:00, total=0:00:00, wall=19:32 EST
    test loss:0.663 │ 100.00% of 64x4... rate=15867.59 Hz, eta=0:00:00, total=0:00:00, wall=19:32 EST
    INFO: === finish epoch 1 / 3 : deploy_demo ===
    epoch lr:0.07 │ vloss is unevaluated 2/3... rate=1.04 Hz, eta=0:00:00, total=0:00:01, wall=19:32 EST
    train loss:0.686 │ 100.00% of 64x8... rate=2743.56 Hz, eta=0:00:00, total=0:00:00, wall=19:32 EST
    test loss:0.636 │ 100.00% of 64x4... rate=14332.63 Hz, eta=0:00:00, total=0:00:00, wall=19:32 EST
    INFO: === finish epoch 2 / 3 : deploy_demo ===
    epoch lr:0.1 │ vloss is unevaluated 3/3... rate=1.11 Hz, eta=0:00:00, total=0:00:02, wall=19:32 EST
    INFO: Maximum harn.epoch reached, terminating ...
    INFO:
    INFO: training completed
    INFO: harn.train_dpath = '.../.cache/netharn/tests/deploy/fit/runs/deploy_demo/onnxqaww'
    INFO: harn.nice_dpath  = '.../.cache/netharn/tests/deploy/fit/nice/deploy_demo'
    INFO: view tensorboard results for this run via:
        tensorboard --logdir ~/.cache/netharn/tests/deploy/fit/nice
    [DEPLOYER] Deployed zipfpath=.../.cache/netharn/tests/deploy/fit/runs/deploy_demo/onnxqaww/deploy_ToyNet2d_onnxqaww_002_TXZBYL.zip
    INFO: wrote single-file deployment to: '.../.cache/netharn/tests/deploy/fit/runs/deploy_demo/onnxqaww/deploy_ToyNet2d_onnxqaww_002_TXZBYL.zip'
    INFO: exiting fit harness.
    Out[1]: '.../.cache/netharn/tests/deploy/fit/runs/deploy_demo/onnxqaww/deploy_ToyNet2d_onnxqaww_002_TXZBYL.zip'
    >>> #
    >>> ##########################################
    >>> print('--- STEP 2: DEPLOY THE MODEL ---')
    >>> # First we export the model topology to a standalone file
    >>> # (Note: this step is done automatically in `harn.run`, but we do
    >>> #  it again here for demo purposes)
    >>> from netharn.export import exporter
    >>> topo_fpath = exporter.export_model_code(harn.train_dpath, harn.hyper.model_cls, harn.hyper.model_params)
    >>> # Now create an instance of deployed model that points to the
    >>> # Training dpath. (Note the directory structure setup by netharn is
    >>> # itself a deployment, it just has multiple files)
    >>> deployer = DeployedModel(harn.train_dpath)
    >>> # Use the DeployedModel to package the imporant info in train_dpath
    >>> # into a standalone zipfile.
    >>> zip_fpath = deployer.package()
    >>> print('We exported the topology to: {!r}'.format(topo_fpath))
    >>> print('We exported the topology+weights to: {!r}'.format(zip_fpath))
    --- STEP 2: DEPLOY THE MODEL ---
    We exported the topology to: '...tests/deploy/fit/runs/deploy_demo/onnxqaww/ToyNet2d_2a3f49.py'
    We exported the topology+weights to: '...tests/deploy/fit/runs/deploy_demo/onnxqaww/deploy_ToyNet2d_onnxqaww_002_HVWCGI.zip'
    >>> #
    >>> #################################################
    >>> print('--- STEP 3: LOAD THE DEPLOYED MODEL ---')
    >>> # Now we can move the zipfile anywhere we want, and we should
    >>> # still be able to load it (depending on how coupled the model is).
    >>> # Create an instance of DeployedModel that points to the zipfile
    >>> # (Note: DeployedModel is used to both package and load models)
    >>> loader = DeployedModel(zip_fpath)
    >>> model = loader.load_model()
    >>> # This model is now loaded with the corret weights.
    >>> # You can use it as normal.
    >>> model.eval()
    >>> images = harn._demo_batch(0)['input'][0:1]
    >>> outputs = model(images)
    >>> print('outputs = {!r}'.format(outputs))
    >>> # Not that the loaded model is independent of harn.model
    >>> print('model.__module__ = {!r}'.format(model.__module__))
    >>> print('harn.model.module.__module__ = {!r}'.format(harn.model.module.__module__))
    --- STEP 3: LOAD THE DEPLOYED MODEL ---
    outputs = tensor([[0.4105, 0.5895]], grad_fn=<SoftmaxBackward>)
    model.__module__ = 'deploy_ToyNet2d_onnxqaww_002_HVWCGI/ToyNet2d_2a3f49'
    harn.model.module.__module__ = 'netharn.models.toynet'
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import glob
import json
import six
import ubelt as ub
# import warnings
import zipfile
import os
from os.path import exists
from os.path import isdir
from os.path import join
from os.path import relpath

__all__ = ['DeployedModel']

if six.PY2:
    FileNotFoundError = OSError


def existing_snapshots(train_dpath):
    # NOTE: Specific to netharn directory structure
    import parse
    snapshot_dpath = join(train_dpath, 'torch_snapshots/')
    prev_states = sorted(glob.glob(join(snapshot_dpath, '_epoch_*.pt')))
    snapshots = {parse.parse('{}_epoch_{num:d}.pt', path).named['num']: path
                 for path in prev_states}
    return snapshots


def find_best_snapshot(train_dpath):
    """
    Returns snapshot written by monitor if available otherwise takes the last
    one.
    """
    # NOTE: Specific to netharn directory structure
    # Netharn should populate best_snapshot.pt if there is a validation set.
    # Other names are to support older codebases.
    expected_names = [
        'best_snapshot.pt',
        'best_snapshot2.pt',
        'final_snapshot.pt',
        'deploy_snapshot.pt',
    ]
    for snap_fname in expected_names:
        snap_fpath = join(train_dpath, snap_fname)
        if exists(snap_fpath):
            break

    if not exists(snap_fpath):
        snap_fpath = None

    if not snap_fpath:
        epoch_to_fpath = existing_snapshots(train_dpath)
        if epoch_to_fpath:
            snap_fpath = epoch_to_fpath[max(epoch_to_fpath)]
    return snap_fpath


def unpack_model_info(path):
    """
    return paths to the most relevant files in a zip or path deployment.

    If path is not a zipfile, this function expects a netharn fit directory
    structure.

    Args:
        path (PathLike): either a zip deployment or train_dpath.
            Preferably this is a zip deployment file or a path to an unzipped
            deploy file. If this is a train_dpath, then it should at least
            contain a model topology py file and snapshot pt file, otherwise
            subsequent usage will likely fail.
    """
    info = {
        'train_info_fpath': None,
        'snap_fpath': None,
        'model_fpath': None,

        # TODO: need to rename and allow a list of arbitrary files
        'glance': [],  # a list of files in the glance directory
    }
    def populate(root, fpaths):
        # TODO: make more robust
        for fpath in fpaths:
            # FIXME: make this more general and robust
            if fpath.endswith('.json'):
                info['train_info_fpath'] = join(root, fpath)
            if fpath.endswith('.pt'):
                info['snap_fpath'] = join(root, fpath)
            if fpath.endswith('.py'):
                new_fpath = join(root, fpath)
                if info['model_fpath'] is not None:
                    try:
                        # Try to take the most recent path if possible.
                        # This will fail if the file is in a zipfile
                        # (because we should not package multiple models)
                        cur_time = os.stat(info['model_fpath']).st_mtime
                        new_time = os.stat(new_fpath).st_mtime
                        if new_time < cur_time:
                            continue  # Keep the current path
                    except OSError:
                        raise Exception(
                            'Multiple model paths! {} and {}'.format(
                                info['model_fpath'], fpath))
                info['model_fpath'] = new_fpath
            # TODO: make including arbitrary files easier
            if fpath.startswith(('glance/', 'glance\\')):
                info['glance'].append(join(root, fpath))

    if path.endswith('.zip'):
        zipfpath = path
        myzip = zipfile.ZipFile(zipfpath, 'r')
        with zipfile.ZipFile(zipfpath, 'r') as myzip:
            populate(zipfpath, (f.filename for f in myzip.filelist))

    elif exists(path) and isdir(path):
        # Populate core files
        populate(path, os.listdir(path))
        # Populate extra glanceable files
        populate(path, [
            relpath(p, path) for p in glob.glob(join(path, 'glance/*'))])
        # If there are no snapshots in the root directory, then
        # use the latest snapshot from the torch_snapshots dir
        if info['snap_fpath'] is None:
            info['snap_fpath'] = find_best_snapshot(path)

    else:
        raise ValueError('cannot unpack model ' + path)
    return info


def _make_package_name2(info):
    """
    Construct a unique and descriptive name for the deployment
    """
    snap_fpath = info['snap_fpath']
    model_fpath = info['model_fpath']
    train_info_fpath = info['train_info_fpath']

    if train_info_fpath and exists(train_info_fpath):
        train_info = json.load(open(train_info_fpath, 'r'))
        model_name = train_info['hyper']['model'][0].split('.')[-1]
        train_hash = ub.hash_data(train_info['train_id'], hasher='sha512',
                                  base='abc', types=True)[0:8]
    else:
        model_name = os.path.splitext(os.path.basename(model_fpath))[0]
        train_hash = 'UNKNOWN-TRAINID'
        print('WARNING: Train info metadata does not exist')

    try:
        # netharn models contain epoch info in the weights file
        import torch
        state = torch.load(snap_fpath,
                           map_location=lambda storage, location: storage)
        epoch = '{:03d}'.format(state['epoch'])
    except Exception:
        epoch = 'UNKNOWN-EPOCH'

    weights_hash = ub.hash_file(snap_fpath, base='abc',
                                hasher='sha512')[0:6].upper()

    deploy_name = 'deploy_{model}_{trainid}_{epoch}_{weights}'.format(
        model=model_name, trainid=train_hash, epoch=epoch,
        weights=weights_hash)
    return deploy_name


def _package_deploy2(dpath, info, name=None):
    """
    Combine the model, weights, and info files into a single deployable file

    Args:
        dpath (PathLike): where to dump the deployment
        info (Dict): containing model_fpath and snap_fpath and optionally
            train_info_fpath and glance, which is a list of extra files.
        name (str, default=None): the name of the zipfile to deploy to.
            If not specified, one will be constructed.

    Ignore:
        dpath = '/home/joncrall/.cache/netharn/tests/_package_custom'
        path = '/home/joncrall/work/opir/fit/nice/_Sim3-kw6-99-finetune_ML3D_BEST_2018-9-20_LR1e-4_f2_vel0.0_hn0.25_bs64_nr5.0'
        info = unpack_model_info(path)
        zipfpath = _package_deploy2(dpath, info)


    """
    model_fpath = info['model_fpath']
    snap_fpath = info['snap_fpath']
    train_info_fpath = info.get('train_info_fpath', None)

    if not snap_fpath:
        raise FileNotFoundError('No weights are associated with the model')

    if name is None:
        deploy_name = _make_package_name2(info)
        deploy_fname = deploy_name + '.zip'
    else:
        if not name.endswith('.zip'):
            raise ValueError('The deployed package name must end in .zip')
        deploy_name = os.path.splitext(name)[0]
        deploy_fname = name

    def zwrite(myzip, fpath, fname=None):
        if fname is None:
            fname = relpath(fpath, dpath)
        myzip.write(fpath, arcname=join(deploy_name, fname))

    zipfpath = join(dpath, deploy_fname)
    with zipfile.ZipFile(zipfpath, 'w') as myzip:
        if train_info_fpath and exists(train_info_fpath):
            zwrite(myzip, train_info_fpath, fname='train_info.json')
        zwrite(myzip, snap_fpath, fname='deploy_snapshot.pt')
        zwrite(myzip, model_fpath, fname=os.path.basename(model_fpath))
        # Add some quick glanceable info
        for p in info.get('glance', []):
            zwrite(myzip, p, fname=join('glance', os.path.basename(p)))
        # for bestacc_fpath in glob.glob(join(train_dpath, 'best_epoch_*')):
        #     zwrite(myzip, bestacc_fpath)
        # for p in glob.glob(join(train_dpath, 'glance/*')):
        #     zwrite(myzip, p)
    print('[DEPLOYER] Deployed zipfpath={}'.format(zipfpath))
    return zipfpath


class DeployedModel(ub.NiceRepr):
    """
    Can setup an initializer and model from a deployed zipfile or a train path

    CommandLine:
        xdoctest -m netharn.export.deployer DeployedModel

    Example:
        >>> # Test the train folder as the model deployment
        >>> train_dpath = _demodata_trained_dpath()
        >>> self = DeployedModel(train_dpath)
        >>> model_ = self.model_definition()
        >>> initializer_ = self.initializer_definition()
        >>> model = model_[0](**model_[1])
        >>> assert initializer_[1].get('fpath', None) is not None, 'initializer isnt setup correctly'
        >>> initializer = initializer_[0](**initializer_[1])
        >>> initializer(model)
        ...
        >>> print('model.__module__ = {!r}'.format(model.__module__))

        # >>> if six.PY3:
        # ...     assert model.__module__ == 'ToyNet2d_2a3f49'
        # ... else:
        # ...     assert model.__module__ == 'ToyNet2d_d573a3'

    Example:
        >>> # Test the zip file as the model deployment
        >>> zip_fpath = _demodata_zip_fpath()
        >>> self = DeployedModel(zip_fpath)
        >>> model_ = self.model_definition()
        >>> initializer_ = self.initializer_definition()
        >>> model = model_[0](**model_[1])
        >>> assert initializer_[1].get('fpath', None) is not None, 'initializer isnt setup correctly'
        >>> initializer = initializer_[0](**initializer_[1])
        >>> initializer(model)
        ...
        >>> # NOTE: the module name should be consistent, but due to
        >>> # small library changes it often changes, so we are permissive
        >>> # with this got/want test
        >>> print('model.__module__ = {!r}'.format(model.__module__))
        model.__module__ = 'deploy_ToyNet2d_..._.../ToyNet2d_...'

        model.__module__ = 'deploy_ToyNet2d_mhuhweia_000_.../ToyNet2d_...'

        model.__module__ = 'deploy_ToyNet2d_rljhgepw_000_.../ToyNet2d_2a3f49'
    """
    def __init__(self, path):
        self.path = path
        self._model = None
        self._info = None

    @classmethod
    def custom(DeployedModel, snap_fpath, model, initkw=None, train_info_fpath=None):
        """
        Create a deployed model even if the model wasnt trained with FitHarn

        This just requires specifying a bit more information, which FitHarn
        would have tracked.

        Args:
            snap_fpath (PathLike):
                path to the exported (snapshot) weights file

            model (PathLike or nn.Module): can either be
                (1) a path to model topology (created via `export_model_code`)
                (2) the model class or an instance of the class

            initkw (Dict): if model is a class or instance, then
                you must pass the keyword arguments used to construct it.

            train_info_fpath (PathLike, optional):
                path to a json file containing additional training metadata

        Example:
            >>> # Setup raw components
            >>> train_dpath = _demodata_trained_dpath()
            >>> deployed = DeployedModel(train_dpath)
            >>> snap_fpath = deployed.info['snap_fpath']
            >>> model, initkw = deployed.model_definition()
            >>> train_info_fpath = deployed.info['train_info_fpath']
            >>> # Past raw components to custom
            >>> self = DeployedModel.custom(snap_fpath, model, initkw)
            >>> dpath = ub.ensure_app_cache_dir('netharn', 'tests/_package_custom')
            >>> self.package(dpath)
        """
        if isinstance(model, six.string_types):
            model_fpath = model
            if initkw is not None:
                raise ValueError('initkw not used when model is a path')
        else:
            import tempfile
            from netharn.export import exporter
            dpath = tempfile.mkdtemp()
            model_fpath = exporter.export_model_code(dpath, model, initkw=initkw)

        _info = {
            'model_fpath': model_fpath,
            'snap_fpath': snap_fpath,
            'train_info_fpath': train_info_fpath,
        }
        self = DeployedModel(None)
        self._info = _info
        return self

    def __nice__(self):
        return self.__json__()

    def __json__(self):
        if self.path is None:
            if self._info:
                return ub.repr2(self._info, nl=0)
        else:
            return self.path

    def package(self, dpath=None, name=None):
        """
        If self.path is a directory, packages important info into a deployable
        zipfile.

        Args:
            dpath (PathLike, optional): directory to dump your packaged model.
                If not specified, it uses the netharn train_dpath if available.
            name (str, default=None): the name of the zipfile to deploy to.
                If not specified, one will be constructed.

        Returns:
            PathLike: path to single-file deployment
        """
        if dpath is None:
            if self.path is None:
                raise ValueError('Must specify dpath for custom deployments')
            else:
                if self.path.endswith('.zip'):
                    raise Exception('Deployed model is already a package')
                dpath = self.path

        zip_fpath = _package_deploy2(dpath, self.info, name=name)
        return zip_fpath

    @property
    def info(self):
        if self._info is None:
            self._info = self.unpack_info()
        return self._info

    def unpack_info(self):
        return unpack_model_info(self.path)

    def model_definition(self):
        model_fpath = self.info['model_fpath']
        module = ub.import_module_from_path(model_fpath)

        export_version = getattr(module, '__pt_export_version__', '0')
        export_version = list(map(int, export_version.split('.')))
        if export_version >= [0, 2, 0]:
            model_cls = module.get_model_cls()
            initkw = module.get_initkw()
        else:
            # Hack to get information from older versions of pytorch_export
            import inspect
            from xdoctest import static_analysis
            print('Hacking to grab model_cls and initkw')
            model = module.make()
            model_cls = model.__class__
            source = inspect.getsource(module.make)
            print(source)
            initkw = static_analysis.parse_static_value('initkw', source=source)
            # Try to reconstruct initkw
        model_ = (model_cls, initkw)
        return model_

    def initializer_definition(self):
        import netharn as nh
        initializer_ = (nh.initializers.Pretrained,
                        {'fpath': self.info['snap_fpath']})
        return initializer_

    def train_info(self):
        import netharn as nh
        train_info_fpath = self.info.get('train_info_fpath', None)
        if train_info_fpath is not None:
            train_info = json.load(nh.util.zopen(train_info_fpath, 'r'))
        else:
            train_info = None
        return train_info

    def load_model(self):
        if self._model is not None:
            return self._model

        model_cls, model_kw = self.model_definition()
        model = model_cls(**model_kw)

        if True:
            # Always load models onto the CPU first
            # import netharn as nh
            model = model.to('cpu')
            # devices = {k: item.device for k, item in model.state_dict().items()}
            # nh.XPU.from_data(model)

        # TODO: load directly from instead of using initializer self.info['snap_fpath']?
        # Actually we can't because we lose the zopen stuff. Its probably ok
        # To depend on netharn a little bit.
        # import torch
        # info = self.unpack_info()
        # state_dict = torch.load(self.info['snap_fpath'])
        # model.load_state_dict()

        initializer_ = self.initializer_definition()
        initializer = initializer_[0](**initializer_[1])

        assert model is not None

        initializer(model)
        return model

    @classmethod
    def ensure_mounted_model(cls, deployed, xpu=None, log=print):
        """
        Ensure that a deployed model is loaded and mounted.

        Helper method that can accept either a raw model or packaged deployed
        model is loaded and mounted on a specific XPU. This provides a one line
        solution for applications that may want to ensure that a model is
        mounted and ready for predict. When the model is already mounted this
        is very fast and just passes the data through. If the input is a
        packaged deployed file, then it does the required work to prep the
        model.

        Args:
            deployed (DeployedModel | PathLike | torch.nn.Module):
                either a packed deployed model, a path to a deployed model, or
                an already mounted torch Module.
            xpu (str | XPU): which device to mount on
            log (callable, optional): logging or print function

        Returns:
            Tuple[torch.nn.Module, XPU]:
                the mounted model, and the device it is mounted on.
        """
        import netharn as nh
        import torch
        if not isinstance(xpu, nh.XPU):
            xpu = nh.XPU.coerce(xpu)

        if isinstance(deployed, six.string_types):
            deployed = nh.export.DeployedModel(deployed)

        if isinstance(deployed, torch.nn.Module):
            # User passed in the model directly
            model = deployed
            try:
                if xpu != nh.XPU.from_data(model):
                    log('Re-Mount model on {}'.format(xpu))
                    model = xpu.mount(model)
            except Exception:
                log('Re-Mount model on {}'.format(xpu))
                model = xpu.mount(model)
        elif isinstance(deployed, nh.export.DeployedModel):
            model = deployed.load_model()
            log('Mount {} on {}'.format(deployed, xpu))
            model = xpu.mount(model)
        else:
            raise TypeError('Unable to ensure {!r} as a mounted model'.format(
                deployed))

        return model, xpu

    @classmethod
    def coerce(DeployedModel, arg):
        """
        Attempt to coerce the argument into a deployed model.

        Args:
            arg (DeployedModel | PathLike | torch.nn.Module) : can be:
                (1) a DeployedModel object
                (2) a path to a deploy file
                (3) a live pytorch module
                (4) a path to a .pt file in a netharn train snapshot directory.

        Returns:
            DeployedModel
        """
        from os.path import dirname
        import torch
        if isinstance(arg, DeployedModel):
            # The input is already a DeployedModel
            deployed = arg
        elif isinstance(arg, torch.nn.Module):
            # The argument is a live pytorch model
            deployed = DeployedModel(None)
            deployed._model = arg
        elif isinstance(arg, six.string_types):
            # handle the case where we are given a weights file
            # use heuristics try and determine model topology
            if arg.endswith('.pt'):
                snap_fpath = arg
                dpath_cands = []
                # Look the pt file's directory for topology and train info
                dpath1 = dirname(snap_fpath)
                dpath_cands = [dpath1]
                # The files might also be in the parent directory
                if not exists(join(dpath1, 'train_info.json')):
                    dpath_cands.append(dirname(dpath1))
                # Search for the files in the candidate directories
                train_info_cands = list(ub.find_path(
                    'train_info.json', path=dpath_cands, exact=True))
                model_cands = list(ub.find_path(
                    '*.py', path=dpath_cands, exact=False))
                if len(model_cands) == 0:
                    raise Exception('Model topology does not exist for {!r}.'.format(arg))
                elif len(model_cands) > 1:
                    raise Exception('Conflicting model topologies for {!r}.'.format(arg))
                else:
                    model_fpath = model_cands[0]
                if len(train_info_cands) == 0:
                    train_info_fpath = None
                elif len(train_info_cands) > 1:
                    raise AssertionError('Conflicting train_info.json files')
                else:
                    train_info_fpath = train_info_cands[0]
                deployed = DeployedModel.custom(snap_fpath, model_fpath,
                                                train_info_fpath=train_info_fpath)
            else:
                # Assume we have a netharn deploy path
                deployed = DeployedModel(arg)
        else:
            # Unhandled case
            raise TypeError(type(arg))
        return deployed


def _demodata_zip_fpath():
    zip_path = DeployedModel(_demodata_trained_dpath()).package()
    return zip_path


def _demodata_toy_harn():
    # This will train a toy model with toy data using netharn
    import netharn as nh
    hyper = nh.HyperParams(**{
        'workdir'     : ub.ensure_app_cache_dir('netharn/tests/deploy'),
        'nice'        : 'deploy_demo_static',
        'xpu'         : nh.XPU.coerce('cpu'),
        'datasets'    : {'train': nh.data.ToyData2d(size=3, rng=0)},
        'loaders'     : {'batch_size': 64},
        'model'       : (nh.models.ToyNet2d, {}),
        'optimizer'   : (nh.optimizers.SGD, {'lr': 0.0001}),
        'criterion'   : (nh.criterions.FocalLoss, {}),
        'initializer' : (nh.initializers.KaimingNormal, {}),
        'monitor'     : (nh.Monitor, {'max_epoch': 1}),
    })
    harn = nh.FitHarn(hyper)
    harn.preferences['use_tensorboard'] = False
    return harn


def _demodata_trained_dpath():
    harn = _demodata_toy_harn()
    harn.run()  # TODO: make this run faster if we don't need to rerun
    if len(list(glob.glob(join(harn.train_dpath, '*.py')))) > 1:
        # If multiple models are deployed some hash changed. Need to reset
        harn.initialize(reset='delete')
        harn.run()  # don't relearn if we already finished this one
    return harn.train_dpath


if __name__ == '__main__':
    """
    CommandLine:
        xdoctest -m netharn.export.deployer all
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# NOTE: pip install -U --pre h5py
from __future__ import absolute_import, division, print_function
import sys
from os.path import dirname
from setuptools import find_packages
from os.path import exists
from os.path import join
import glob
import os
from setuptools import setup
# from skbuild import setup


def parse_version(fpath):
    """
    Statically parse the version number from a python file
    """
    import ast
    if not exists(fpath):
        raise ValueError('fpath={!r} does not exist'.format(fpath))
    with open(fpath, 'r') as file_:
        sourcecode = file_.read()
    pt = ast.parse(sourcecode)
    class VersionVisitor(ast.NodeVisitor):
        def visit_Assign(self, node):
            for target in node.targets:
                if getattr(target, 'id', None) == '__version__':
                    self.version = node.value.s
    visitor = VersionVisitor()
    visitor.visit(pt)
    return visitor.version


def parse_description():
    """
    Parse the description in the README file

    CommandLine:
        pandoc --from=markdown --to=rst --output=README.rst README.md
        python -c "import setup; print(setup.parse_description())"
    """
    from os.path import dirname, join, exists
    readme_fpath = join(dirname(__file__), 'README.rst')
    # This breaks on pip install, so check that it exists.
    if exists(readme_fpath):
        try:
            with open(readme_fpath, 'r') as f:
                text = f.read()
            return text
        except Exception as ex:
            import warnings
            warnings.warn('unable to parse existing readme: {!r}'.format(ex))
    return ''


def parse_requirements(fname='requirements.txt', with_version=False):
    """
    Parse the package dependencies listed in a requirements file but strips
    specific versioning information.

    Args:
        fname (str): path to requirements file
        with_version (bool, default=False): if true include version specs

    Returns:
        List[str]: list of requirements items

    References:
        https://pip.readthedocs.io/en/1.1/requirements.html

    CommandLine:
        python -c "import setup; print(setup.parse_requirements())"
        python -c "import setup; print(chr(10).join(setup.parse_requirements(with_version=True)))"
    """
    from os.path import exists
    import re
    require_fpath = fname

    def parse_line(line, base='.'):
        """
        Parse information from a line in a requirements text file
        """
        if line.startswith(('-f ', '--find-links ', '--index-url ')):
            import warnings
            warnings.warn(
                'requirements file specified alternative index urls, but '
                'there is currently no way to support this in setuptools')
        elif line.startswith('-r '):
            # Allow specifying requirements in other files
            new_fname = line.split(' ')[1]
            new_fpath = join(base, new_fname)
            for info in parse_require_file(new_fpath):
                yield info
        else:
            info = {'line': line}
            if line.startswith('-e '):
                info['package'] = line.split('#egg=')[1]
            else:
                # Remove versioning from the package
                pat = '(' + '|'.join(['>=', '==', '>']) + ')'
                parts = re.split(pat, line, maxsplit=1)
                parts = [p.strip() for p in parts]

                info['package'] = parts[0]
                if len(parts) > 1:
                    op, rest = parts[1:]
                    if ';' in rest:
                        # Handle platform specific dependencies
                        # http://setuptools.readthedocs.io/en/latest/setuptools.html#declaring-platform-specific-dependencies
                        version, platform_deps = map(str.strip, rest.split(';'))
                        info['platform_deps'] = platform_deps
                    else:
                        version = rest  # NOQA
                    info['version'] = (op, version)
            yield info

    def parse_require_file(fpath):
        base = dirname(fpath)
        with open(fpath, 'r') as f:
            for line in f.readlines():
                line = line.strip()
                if line and not line.startswith('#'):
                    for info in parse_line(line, base):
                        yield info

    def gen_packages_items():
        if exists(require_fpath):
            for info in parse_require_file(require_fpath):
                parts = [info['package']]
                if with_version and 'version' in info:
                    parts.extend(info['version'])
                if not sys.version.startswith('3.4'):
                    # apparently package_deps are broken in 3.4
                    platform_deps = info.get('platform_deps')
                    if platform_deps is not None:
                        parts.append(';' + platform_deps)
                item = ''.join(parts)
                yield item

    packages = list(gen_packages_items())
    return packages


def clean_repo(repodir, modname, rel_paths=[]):
    """
    repodir = ub.expandpath('~/code/netharn/')
    modname = 'netharn'
    rel_paths = [
        'netharn/util/nms/cpu_nms.c',
        'netharn/util/nms/cpu_nms.c',
        'netharn/util/nms/cpu_nms.cpp',
        'netharn/util/nms/cython_boxes.c',
        'netharn/util/nms/cython_boxes.html',
    ]
    """
    print('cleaning repo: {}/{}'.format(repodir, modname))
    toremove = []
    for root, dnames, fnames in os.walk(repodir):

        if os.path.basename(root) == modname + '.egg-info':
            toremove.append(root)
            del dnames[:]

        if os.path.basename(root) == '__pycache__':
            toremove.append(root)
            del dnames[:]

        if os.path.basename(root) == '_ext':
            # Remove torch extensions
            toremove.append(root)
            del dnames[:]

        if os.path.basename(root) == 'build':
            # Remove python c extensions
            if len(dnames) == 1 and dnames[0].startswith('temp.'):
                toremove.append(root)
                del dnames[:]

        # Remove simple pyx inplace extensions
        for fname in fnames:
            if fname.endswith('.pyc'):
                toremove.append(join(root, fname))
            if fname.endswith(('.so', '.c', '.o')):
                if fname.split('.')[0] + '.pyx' in fnames:
                    toremove.append(join(root, fname))

    def enqueue(d):
        if exists(d) and d not in toremove:
            toremove.append(d)

    import six
    if six.PY2:
        abs_paths = [join(repodir, p) for pat in rel_paths
                     for p in glob.glob(pat)]
    else:
        abs_paths = [join(repodir, p) for pat in rel_paths
                     for p in glob.glob(pat, recursive=True)]
    for abs_path in abs_paths:
        enqueue(abs_path)

    import ubelt as ub
    for dpath in toremove:
        # print('Removing dpath = {!r}'.format(dpath))
        ub.delete(dpath, verbose=1)


def clean():
    """
    __file__ = ub.expandpath('~/code/netharn/setup.py')
    """
    modname = 'netharn'
    repodir = dirname(__file__)
    rel_paths = [
        'htmlcov',
        '_skbuild',
        '_build_wheel',
        'netharn.egg-info',
        'dist',
        'build',
        '**/*.pyc',
        'profile*'
        'pip-wheel-metadata',
    ]
    clean_repo(repodir, modname, rel_paths)


VERSION = version = parse_version('netharn/__init__.py')  # needs to be a global var for git tags
NAME = 'netharn'

if __name__ == '__main__':
    if 'clean' in sys.argv:
        clean()
        # sys.exit(0)

    setup(
        name=NAME,
        version=VERSION,
        author='Jon Crall',
        author_email='jon.crall@kitware.com',
        url='https://gitlab.kitware.com/computer-vision/netharn',
        description='Train and deploy pytorch models',
        long_description=parse_description(),
        long_description_content_type='text/x-rst',
        install_requires=parse_requirements('requirements/runtime.txt'),
        extras_require={
            'all': parse_requirements('requirements.txt'),
            'optional': parse_requirements('requirements/optional.txt'),
            'tests': parse_requirements('requirements/tests.txt'),
        },
        packages=find_packages(include='netharn.*'),
        license='Apache 2',
        classifiers=[
            # List of classifiers available at:
            # https://pypi.python.org/pypi?%3Aaction=list_classifiers
            'Development Status :: 4 - Beta',
            'Intended Audience :: Developers',
            'Intended Audience :: Science/Research',
            'Topic :: Scientific/Engineering',
            'Topic :: Scientific/Engineering :: Artificial Intelligence',
            'Topic :: Software Development',
            'Topic :: Software Development :: Libraries :: Python Modules',
            'Topic :: Utilities',
            # This should be interpreted as Apache License v2.0
            'License :: OSI Approved :: Apache Software License',
            # Supported Python versions
            'Programming Language :: Python :: 3',
        ],
    )
